#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import time
import sys
import io
from multiprocessing import Process, Queue
import logging
from ibm_watson_machine_learning.experiments import Experiments
from preparation_and_cleaning import *


@unittest.skip("Skip experiments.")
class TestWMLClientWithExperiment(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    scoring_url = None
    definition_1_uid = None
    definition_1_url = None
    definition_2_uid = None
    definition_2_url = None
    trained_model_uid = None
    experiment_uid = None
    experiment_run_uid = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithExperiment.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.cos_resource = get_cos_resource()
        self.bucket_names = prepare_cos(self.cos_resource)

    @classmethod
    def tearDownClass(self):
        clean_cos(self.cos_resource, self.bucket_names)

    def test_01_service_instance_details(self):
        TestWMLClientWithExperiment.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithExperiment.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithExperiment.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_save_definition1(self):
        TestWMLClientWithExperiment.logger.info("Save model definition ...")

        self.client.repository.DefinitionMetaNames.show()

        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "my_training_definition",
            self.client.repository.DefinitionMetaNames.DESCRIPTION: "my_description",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "tensorflow",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "1.5",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python3 tensorflow_mnist_softmax.py --trainingIters 20"
            }

        model_content_path = '../../svt/artifacts/tf-softmax-model.zip'
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithExperiment.definition_1_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithExperiment.definition_1_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithExperiment.logger.info("Saved model definition uid: " + str(TestWMLClientWithExperiment.definition_1_uid))

    def test_03_save_definition2(self):
        TestWMLClientWithExperiment.logger.info("Save model definition ...")
        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "my_training_definition",
            self.client.repository.DefinitionMetaNames.DESCRIPTION: "my_description",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "tensorflow",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "1.5",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python3 tensorflow_mnist_softmax.py --trainingIters 20"
            }

        model_content_path = '../../svt/artifacts/tf-softmax-model.zip'
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithExperiment.definition_2_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithExperiment.definition_2_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithExperiment.logger.info("Saved model definition uid: " + str(TestWMLClientWithExperiment.definition_2_uid))

    def test_04_get_definition_details(self):
        TestWMLClientWithExperiment.logger.info("Getting definition details ...")
        details_1 = self.client.repository.get_definition_details(TestWMLClientWithExperiment.definition_1_uid)
        TestWMLClientWithExperiment.logger.info(details_1)
        self.assertTrue('my_training_definition' in str(details_1))

        details_2 = self.client.repository.get_definition_details(TestWMLClientWithExperiment.definition_2_uid)
        TestWMLClientWithExperiment.logger.info(details_2)
        self.assertTrue('my_training_definition' in str(details_2))

    def test_05_save_experiment(self):
        metadata = {
                    self.client.repository.ExperimentMetaNames.NAME: "xxx",
                    self.client.repository.ExperimentMetaNames.AUTHOR_EMAIL: "js@js.com",
                    self.client.repository.ExperimentMetaNames.EVALUATION_METHOD: "binary",
                    self.client.repository.ExperimentMetaNames.EVALUATION_METRICS: [],
                    self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: get_cos_training_data_reference(self.bucket_names),
                    self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(self.bucket_names),
                    self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                        {
                            "name": "mnist_nn",
                            "training_definition_url": TestWMLClientWithExperiment.definition_1_url,
                            "compute_configuration": {"name": "k80"}
                        },
                        {
                            "name": "mnist_cnn",
                            "training_definition_url": TestWMLClientWithExperiment.definition_2_url,
                            "compute_configuration": {"name": "k80"}
                        },
                    ]
                }

        TestWMLClientWithExperiment.logger.info(get_cos_training_data_reference(self.bucket_names))
        TestWMLClientWithExperiment.logger.info(get_cos_training_results_reference(self.bucket_names))
        experiment_details = self.client.repository.store_experiment(meta_props=metadata)

        TestWMLClientWithExperiment.experiment_uid = self.client.repository.get_experiment_uid(experiment_details)

        experiment_specific_details = self.client.repository.get_experiment_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(experiment_specific_details))

    def test_06_update_experiment(self):
        metadata = {
            self.client.repository.ExperimentMetaNames.NAME: "my_experiment",
            self.client.repository.ExperimentMetaNames.DESCRIPTION: "mnist best model",
            self.client.repository.ExperimentMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.ExperimentMetaNames.EVALUATION_METHOD: "multiclass",
            self.client.repository.ExperimentMetaNames.EVALUATION_METRICS: ["accuracy"],
            self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: get_cos_training_data_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                {
                    "name": "mnist_nn",
                    "training_definition_url": TestWMLClientWithExperiment.definition_1_url,
                    "compute_configuration": {"name": self.client.training.ConfigurationMetaNames._COMPUTE_CONFIGURATION_DEFAULT}
                },
                {
                    "name": "mnist_cnn",
                    "training_definition_url": TestWMLClientWithExperiment.definition_2_url,
                    "compute_configuration": {"name": self.client.training.ConfigurationMetaNames._COMPUTE_CONFIGURATION_DEFAULT}
                },
            ]
        }

        experiment_details = self.client.repository.update_experiment(TestWMLClientWithExperiment.experiment_uid, metadata)
        self.assertTrue('my_experiment' in str(experiment_details))
        TestWMLClientWithExperiment.logger.info(experiment_details)
        self.assertTrue('xxx' not in str(experiment_details))

    def test_07_get_experiment_details(self):
        details = self.client.repository.get_experiment_details()
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(details))

        details2 = self.client.repository.get_experiment_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(details2))

    def test_08_run_experiment(self):
        created_experiment_run_details = self.client.experiments.run(TestWMLClientWithExperiment.experiment_uid, asynchronous=True)
        self.assertIsNotNone(created_experiment_run_details)
        TestWMLClientWithExperiment.experiment_run_uid = Experiments.get_run_uid(created_experiment_run_details)

    def test_09_monitor_metrics(self):
        queue = Queue()
        process = Process(target=run_monitor, args=(self.client, TestWMLClientWithExperiment.experiment_run_uid, queue))
        process.start()
        process.join(timeout=600)
        process.terminate()
        self.assertFalse(queue.empty())
        self.assertTrue('accuracy' in queue.get())

    def test_10_get_status(self):
        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 60 * 10:
            time.sleep(3)
            status = self.client.experiments.get_status(TestWMLClientWithExperiment.experiment_run_uid)
            if status['state'] == 'completed' or status['state'] == 'error' or status['state'] == 'canceled':
                break

            diff_time = time.time() - start_time
        self.assertIsNotNone(status)
        self.assertTrue(status['state'] == 'completed')

    def test_11_get_all_experiments_runs_details(self):
        details = self.client.experiments.get_details()
        self.assertIsNotNone(details)

    def test_12_get_experiment_run_details(self):
        details = self.client.experiments.get_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertIsNotNone(details)

    def test_13_get_experiment_run_details(self):
        details = self.client.experiments.get_run_details(TestWMLClientWithExperiment.experiment_run_uid)
        self.assertIsNotNone(details)

        self.assertIsNotNone(self.client.experiments.get_training_runs(details))
        self.assertIsNotNone(self.client.experiments.get_training_uids(details))

    def test_14_list_experiments(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs()  # Just to see values.

        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_experiments()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.repository.list_experiments()  # Just to see values.

    def test_15_list_experiment_runs_for_experiment(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs(TestWMLClientWithExperiment.experiment_uid)  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs(TestWMLClientWithExperiment.experiment_uid)  # Just to see values.

    def test_16_delete_experiment_run(self):
        self.client.experiments.delete(TestWMLClientWithExperiment.experiment_run_uid)

    def test_17_list(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_definitions()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.definition_1_uid in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithExperiment.definition_2_uid in captured_output.getvalue())
        self.client.repository.list_definitions()  # Just to see values.

        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_experiments()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.repository.list_experiments()  # Just to see values.

        self.client.repository.list()

    def test_18_delete_experiment(self):
        self.client.repository.delete(TestWMLClientWithExperiment.experiment_uid)

    def test_19_delete_definitions(self):
        self.client.repository.delete(TestWMLClientWithExperiment.definition_1_uid)
        self.client.repository.delete(TestWMLClientWithExperiment.definition_2_uid)

    def test_20_list(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.definition_1_uid not in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithExperiment.definition_2_uid not in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid not in captured_output.getvalue())
        self.client.repository.list()  # Just to see values.


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.utils.autoai.utils import NextRunDetailsGenerator
from ibm_watson_machine_learning.tests.utils import get_wml_credentials


class TestNextRunDetailsGenerator(unittest.TestCase):
    wml_credentials_cp4d = None
    client = None
    project_id = 'dd829201-9d59-4f5a-b0e0-6ea3a88ae66b'
    generator = None
    data = []

    @classmethod
    def setUp(cls) -> None:
        cls.wml_credentials_cp4d = get_wml_credentials('CLOUD_DEV_AM')
        cls.client = APIClient(cls.wml_credentials_cp4d, cls.project_id)
        cls.client.set.default_project(cls.project_id)

    def test_01__initialize_generator_class(self):
        details = self.client.training.get_details(limit=1)
        self.data.extend(details['resources'])
        TestNextRunDetailsGenerator.generator = NextRunDetailsGenerator(wml_client=self.client,
                                                                        href=details['next']['href'])

        self.assertIsInstance(self.generator.wml_client, APIClient, msg="wml client not set")
        self.assertEqual(self.generator.href, details['next']['href'], msg="href is incorrect")

    def test_02__generate_next_run_details(self):
        for entry in self.generator:
            self.data.extend(entry)
        self.assertGreater(1, len(self.data), msg="There is only one or 0 runs, or generator works in incorrect way.")


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging, re

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, PipelineArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.api_client import ApiException
from ibm_watson_machine_learning.libs.repo.swagger_client.models import TagRepository,MlAssetsCreateExperimentInput, MlAssetsCreateModelInput, \
    MlAssetsCreateModelOutput , MlAssetsCreateExperimentOutput

from ibm_watson_machine_learning.libs.repo.util import  Json2ObjectMapper
import json
from .experiment_adapter import ExperimentAdapter
from ibm_watson_machine_learning.libs.repo.swagger_client.models.author_repository import AuthorRepository
from ibm_watson_machine_learning.libs.repo.swagger_client.models.framework_output_repository import FrameworkOutputRepository
from ibm_watson_machine_learning.libs.repo.swagger_client.models.connection_object_with_name_repository import ConnectionObjectWithNameRepository
from ibm_watson_machine_learning.libs.repo.swagger_client.models.ml_assets_create_experiment_input import MlAssetsCreateExperimentInput
from ibm_watson_machine_learning.libs.repo.swagger_client.models.array_data_input_repository import  ArrayDataInputRepository
from ibm_watson_machine_learning.libs.repo.swagger_client.models.framework_output_repository_runtimes import  FrameworkOutputRepositoryRuntimes
from ibm_watson_machine_learning.libs.repo.swagger_client.models.framework_output_repository_libraries import  FrameworkOutputRepositoryLibraries
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.generic_archive_pipeline_model_artifact import GenericArchivePipelineModelArtifact

logger = logging.getLogger('ExperimentCollection')


class ExperimentCollection:
    """
    Client operating on experiments in repository service.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLRepositoryApi repository_api: client connecting to repository rest api
    :param MLRepositoryClient client: high level client used for simplification and argument for constructors
    """
    def __init__(self, base_path, repository_api, client):

        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryApi

        if not isinstance(base_path, str) and not isinstance(base_path, unicode):
            raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

        if not isinstance(repository_api, MLRepositoryApi):
            raise ValueError('Invalid type for repository_api: {}'.format(repository_api.__class__.__name__))

        if not isinstance(client, MLRepositoryClient):
            raise ValueError('Invalid type for client: {}'.format(client.__class__.__name__))

        self.base_path = base_path
        self.repository_api = repository_api
        self.client = client

    def _extract_mlassets_create_experiment_output(self, service_output):
        latest_version = service_output.entity['training_definition_version']
        return ExperimentAdapter(service_output, latest_version, self.client).artifact()

    def all(self, queryMap=None):
        """
        Gets info about all experiments which belong to this user.

        Not complete information is provided by all(). To get detailed information about experiment use get().

        :return: info about experiments
        :rtype: list[ExperimentsArtifact]
        """
        all_experiments = self.repository_api.repository_list_experiments(queryMap)
        list_pipeline_artifact = []
        if all_experiments is not None:
            resr = all_experiments.resources
            for iter1 in resr:
                exper_entity = iter1.entity
                ver_url = iter1.entity['training_definition_version']
                list_pipeline_artifact.append(ExperimentAdapter(iter1, ver_url, self.client).artifact())
            return list_pipeline_artifact
        else:
            return []

    def versions(self, training_definition_id):
        """
        Gets all available versions.

        Not implemented yet.

        :param str training_definition_id: uid used to identify model
        :return: ???
        :rtype: list[str]
        """

        if not isinstance(training_definition_id, str) and  not isinstance(training_definition_id, unicode):
            raise ValueError('Invalid type for training_definition_id: {}'.format(training_definition_id.__class__.__name__))

        experiment_ver_output = self.repository_api.repository_list_experiment_versions(training_definition_id)

        list_pipeline_artifact = [PipelineArtifact]
        if experiment_ver_output is not None:
            resr = experiment_ver_output.resources
            for iter1 in resr:
                exper_entity = iter1.entity
                ver_url = iter1.entity['training_definition_version']
                list_pipeline_artifact.append(ExperimentAdapter(iter1, ver_url, self.client).artifact())
            return list_pipeline_artifact
        else:
            raise ApiException('Pipeline with guid={} not found'.format(training_definition_id))

    def get(self, training_definition_id):

        """
        Gets detailed information about experiment.

        :param str training_definition_id: uid used to identify experiment
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(training_definition_id, str) and  not isinstance(training_definition_id, unicode):
            raise ValueError('Invalid type for training_definition_id: {}'.format(training_definition_id.__class__.__name__))

        experiment_output = self.repository_api.v3_ml_assets_experiments_experiment_id_get(training_definition_id)

        if experiment_output is not None:
            ver_url = experiment_output.entity['training_definition_version']
            return ExperimentAdapter(experiment_output, ver_url, self.client).artifact()
        else:
            raise Exception('Model with guid={} not found'.format(training_definition_id))


    def version(self, training_definition_id, ver):
        """
        Gets experiment version with given training_definition_id and ver

        :param str training_definition_id: uid used to identify experiment
        :param str ver: uid used to identify version of experiment
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(training_definition_id, str) and not isinstance(training_definition_id, unicode):
            raise ValueError('Invalid type for training_definition_id: {}'.format(training_definition_id.__class__.__name__))

        #if not isinstance(ver.encode('ascii'), str):
        if(ver is not None) and not isinstance(ver, str) and not isinstance(ver, unicode):
            raise ValueError('Invalid type for ver: {}'.format(ver.__class__.__name__))

        experiment_output = self.repository_api.repository_get_experiment_version(training_definition_id, ver)

        if experiment_output is not None:
            ver_url = experiment_output.metadata['href']
            return ExperimentAdapter(experiment_output, ver_url, self.client).artifact()
            #return self._extract_experiments_from_output(experiment_output)
        else:
            raise Exception('Model with guid={} not found'.format(training_definition_id))

    def version_from_url(self, artifact_version_url):
        """
        Gets experiment version from given href

        :param str artifact_version_href: href identifying artifact and version
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(artifact_version_url, str) and not isinstance(artifact_version_url, unicode):
            raise ValueError('Invalid type for artifact_version_href: {}'
                             .format(artifact_version_url.__class__.__name__))

        #if artifact_version_url.startswith(self.base_path):

        matchedV2 = re.search(
            '.*/v4/pipelines/([A-Za-z0-9\-]+)$', artifact_version_url)

        matchedV3 = re.search(
            '.*/v4/pipelines/([A-Za-z0-9\-]+)?rev=([A-Za-z0-9\-]+)', artifact_version_url)

        if matchedV2 is not None:
            training_definition_id = matchedV2.group(1)
            version_id = None
            print(training_definition_id, version_id)
            return self.version(training_definition_id, version_id)
        elif matchedV3 is not None:
            training_definition_id = matchedV3.group(1)
            version_id = matchedV3.group(2)
            print (training_definition_id, version_id)
            return self.version(training_definition_id, version_id)
        else:
            raise ValueError('Unexpected artifact version href: {} format'.format(artifact_version_url))
        #else:
        #    raise ValueError('The artifact version href: {} is not within the client host: {}').format(
        #        artifact_version_url,
        #        self.base_path
        #    )

    def remove(self, training_definition_id):
        """
        Removes experiment with given training_definition_id.

        :param str training_definition_id: uid used to identify experiment
        """

        if not isinstance(training_definition_id, str) and not isinstance(training_definition_id, unicode):
            raise ValueError('Invalid type for training_definition_id: {}'.format(training_definition_id.__class__.__name__))

        return self.repository_api.v3_ml_assets_experiments_experiment_id_delete(training_definition_id)

    def save(self, artifact):
        """
        Saves experiment in repository service.

        :param SparkPipelineArtifact artifact: artifact to be saved in the repository service
        :return: saved artifact with changed MetaProps
        :rtype: SparkPipelineArtifact
        """

        logger.debug('Creating a new experiment: {}'.format(artifact.name))

        if not issubclass(type(artifact), PipelineArtifact) and not issubclass(type(artifact),GenericArchivePipelineModelArtifact) :
            raise ValueError('Invalid type for artifact: {}'.format(artifact.__class__.__name__))

        if artifact.meta.prop(MetaNames.PIPELINE_URL) is not None:
            raise ApiException(400, 'Invalid operation: save the same experiment artifact twice')

        try:
            training_definition_id = artifact.uid
            if training_definition_id is None:
                lib_name = artifact.meta.prop(MetaNames.NAME)
                ##create libry
                name = lib_name
                description = "custom libraries for scoring"
                version = "1.0"
                platform = """{"name":"python", "versions":["3.5"]}"""
                lib_metadata = {
                    MetaNames.LIBRARIES.NAME: name,
                    MetaNames.LIBRARIES.VERSION: version,
                    MetaNames.LIBRARIES.DESCRIPTION: description,
                    MetaNames.LIBRARIES.PLATFORM: platform,
                }

               # experiment_input = self._prepare_experiment_input(artifact)

                r = self.repository_api.v3_libraries_create(library_input)
                statuscode = r[1]
                if statuscode != 201:
                    logger.info('Error while creating library: no location header')
                    raise ApiException(404, 'No artifact location')

                experiment_artifact = self._extract_experiments_from_output(r)
                location = r[2].get('Location')

                # location_match = re.search('.*/v3/ml_assets/training_definitions/([A-Za-z0-9\\-]+)', location)

                # if location_match is not None:
                #     training_definition_id = location_match.group(1)
                # else:
                #     logger.info('Error while creating experiment: no location header')
                #     raise ApiException(404, 'No artifact location')
                # artifact_with_guid = artifact._copy(training_definition_id)
                # version_location = experiment_artifact.meta.prop(MetaNames.TRAINING_DEFINITION_VERSION_URL)
                # version_id = experiment_artifact.meta.prop(MetaNames.VERSION)
                experiment_artifact.pipeline_instance = lambda: artifact.ml_pipeline

                if version_location is not None:
                    content_stream = artifact_with_guid.reader().read()
                    self.repository_api.upload_pipeline_version(training_definition_id, version_id, content_stream)
                    content_stream.close()
                    artifact_with_guid.reader().close()
                    return experiment_artifact
                else:
                    logger.info('Error while creating experiment version: no location header')
                    raise ApiException(404, 'No artifact location')
            else:
                raise ApiException(404, 'Pipeline not found')

        except Exception as e:
            logger.info('Error in experiment creation')
            import traceback
            print(traceback.format_exc())
            raise e

    def _extract_experiments_from_output(self, service_output):

        latest_version = service_output[0].entity['training_definition_version']
        return ExperimentAdapter(service_output[0], latest_version, self.client).artifact()

    @staticmethod
    def _prepare_experiment_input(artifact):
        exper_input = MlAssetsCreateExperimentInput()
        exper_input.name = artifact.name
        exper_input.description = artifact.meta.prop(MetaNames.DESCRIPTION)
        frameworkLibraries = artifact.meta.prop(MetaNames.FRAMEWORK_LIBRARIES)
        frlibrary_param_list = []

        if frameworkLibraries is not None:
            frlibraryvalue = json.loads(frameworkLibraries)
            if not issubclass (type(frlibraryvalue), list):
                raise ValueError('Invalid data format for libraries.')
            for iter1 in frlibraryvalue:
                frlibrary_param = FrameworkOutputRepositoryLibraries(
                   iter1.get('name', None),
                   iter1.get('version', None)
                )
                frlibrary_param_list.append(frlibrary_param)
                exper_input.framework = FrameworkOutputRepository(
                      artifact.meta.prop(MetaNames.FRAMEWORK_NAME),
                      artifact.meta.prop(MetaNames.FRAMEWORK_VERSION),
                      libraries=frlibrary_param_list
                      )
        else:
           exper_input.framework = FrameworkOutputRepository(
               artifact.meta.prop(MetaNames.FRAMEWORK_NAME),
               artifact.meta.prop(MetaNames.FRAMEWORK_VERSION)
           )

        exper_input.author = AuthorRepository(
             artifact.meta.prop(MetaNames.AUTHOR_NAME)
        )



        if artifact.meta.prop(MetaNames.TRAINING_DATA_REFERENCE) is not None:
            dataref_list=artifact.meta.prop(MetaNames.TRAINING_DATA_REFERENCE)
            if isinstance(dataref_list, str):
               dataref_list = json.loads(artifact.meta.prop(MetaNames.TRAINING_DATA_REFERENCE))
            training_data_list = []
            if isinstance(dataref_list, list):
                for iter1 in dataref_list:
                    training_data = ConnectionObjectWithNameRepository(
                        iter1.get('name', None),
                        iter1.get('connection', None),
                        iter1.get('source', None)
                    )
                    training_data_list.append(training_data)
            elif isinstance(dataref_list, dict):
                training_data = ConnectionObjectWithNameRepository(
                    dataref_list.get('name', None),
                    dataref_list.get('connection', None),
                    dataref_list.get('source', None)
                )
                training_data_list.append(training_data)
            else:

                raise ApiException(404, 'Pipeline not found')
            exper_input.training_data_reference = training_data_list

        if artifact.meta.prop(MetaNames.TAGS) is not None:
            tags=artifact.meta.prop(MetaNames.TAGS)
            tags_data_list = []
            if isinstance(tags, str):
              tags_list = json.loads(artifact.meta.prop(MetaNames.TAGS))
              if isinstance(tags_list, list):
                for iter1 in tags_list:
                    tags_data = TagRepository()
                    for key in iter1:
                        if key == 'value':
                           tags_data.value= iter1['value']
                        if key == 'description':
                           tags_data.description = iter1['description']
                    tags_data_list.append(tags_data)
              else:
               raise ValueError("Invalid tag Input")
              exper_input.tags =  tags_data_list

        return exper_input

#    @staticmethod
#    def _get_version_input(artifact):
#     return PipelineVersionInput(artifact.meta.prop(MetaNames.PARENT_VERSION))



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class MetaNames(object):
    """
    Holder for constants used by MetaProps.

    Description of keys:

    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |Key name                        |User     |Used with|Used with|Optional|Description                                      |
    |                                |specified|pipeline |model    |        |                                                 |
    +================================+=========+=========+=========+========+=================================================+
    |MetaNames.CREATION_TIME         |No       |Yes      |Yes      |--      |time of creating artifact in repository service  |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.LAST_UPDATED          |No       |Yes      |Yes      |--      |time of last update of this artifact             |
    |                                |         |         |         |        |in repository service                            |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.TRAINING_DATA_REF     |Yes      |No       |Yes      |Yes     |reference to training data for model             |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.LABEL_FIELD           |No       |No       |Yes      |--      |information about model what is the name         |
    |                                |         |         |         |        |of output column (labelCol)                      |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.PARENT_VERSION        |No       |Yes      |Yes      |--      |href to previous version of artifact             |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.VERSION               |No       |Yes      |Yes      |--      |id of version of artifact                        |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.MODEL_METRICS         |--       |No       |Yes      |--      |modelMetrics                                     |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.EVALUATION_METHOD     |--       |No       |Yes      |--      |evaluationMethod                                 |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.EVALUATION_METRICS    |--       |No       |Yes      |--      |evaluationMetrics                                |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.FRAMEWORK_NAME        |Yes      |Yes      |Yes      |--      |Framework type name, used with experiment        |
    |                                |         |         |         |        |and models                                       |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.FRAMEWORK_VERSION     |Yes      |Yes      |Yes      |--      |Framework type version, used with experiments    |
    |                                |         |         |         |        |and models                                       |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.DESCRIPTION           |Yes      |Yes      |Yes      |Yes     |description prepared by user                     |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.MODEL_VERSION_URL     |No       |No       |Yes      |--      |url to version of this model in repository       |
    |                                |         |         |         |        |service, used only with models                   |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.EXPERIMENT_VERSION_URL|No       |Yes      |No       |--      |url to version of this experiment in repository  |
    |                                |         |         |         |        |service, used only with experiment               |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.AUTHOR_NAME           |Yes      |Yes      |Yes      |Yes     |name of author                                   |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.AUTHOR_EMAIL          |Yes      |Yes      |Yes      |Yes     |email of author                                  |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.EXPERIMENT_URL        |No       |Yes      |No       |--      |Url to this experiment in repository             |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.LIBRARIES.URL         |No       |Yes      |No       |--      |Url to this libraries in repository             |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.RUNTIMES.URL          |No       |Yes      |No       |--      |Url to this runtimes in repository             |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+
    |MetaNames.MODEL_URL             |No       |No       |Yes      |--      |Url to this model in repository                  |
    +--------------------------------+---------+---------+---------+--------+-------------------------------------------------+

    """
    CREATION_TIME = "creationTime"
    LAST_UPDATED = "lastUpdated"
    INPUT_DATA_SCHEMA = "inputDataSchema"
    TRAINING_DATA_REFERENCES = "training_data_references"
    RUNTIME = "runtime"
    RUNTIMES = "runtimes"
    FRAMEWORK_RUNTIMES = "framework_runtimes"
    TRAINING_DATA_SCHEMA = "trainingDataSchema"
    LABEL_FIELD = "label_column"
    PARENT_VERSION = "parentVersion"
    VERSION = "version"
    MODEL_METRICS = "modelMetrics"
    EVALUATION_METHOD = "evaluationMethod"
    EVALUATION_METRICS = "evaluationMetrics"
    FRAMEWORK_NAME = "frameworkName"
    FRAMEWORK_VERSION = "frameworkVersion"
    FRAMEWORK_LIBRARIES = "frameworkLibraries"
    DESCRIPTION = "description"
    MODEL_VERSION_URL = "modelVersionUrl"
    TRAINING_DEFINITION_VERSION_URL = "trainingDefinitionVersionUrl"
    AUTHOR_NAME = "authorName"
    AUTHOR_EMAIL = "authorEmail"
    EXPERIMENT_URL = "experimentUrl"
    TRAINING_DEFINITION_URL = "trainingDefinitionUrl"
    MODEL_URL = "modelUrl"
    TRANSFORMED_LABEL_FIELD = "transformed_label"
    CONTENT_STATUS = "contentStatus"
    CONTENT_LOCATION = "contentLocation"
    HYPER_PARAMETERS = "hyperParameters"
    STATUS_URL = "statusUrl"
    TAGS = "tags"
    OUTPUT_DATA_SCHEMA = "outputDataSchema"
    LABEL_VALUES = "labelValues"
    PREDICTION_FIELD = "predictionField"
    PROBABILITY_FIELD = "probabilityField"
    DECODED_LABEL = "decodedLabel"
    CATEGORY = "category"
    PROJECT_UID = "project"
    SPACE_UID = "space"
    CUSTOM = "custom"
    RUNTIME_UID = "runtime"
    PIPELINE_UID = "pipeline"
    TRAINING_LIB_UID = "training_lib"
    TYPE = "type"
    DOMAIN = "domain"
    IMPORT="import"
    SIZE = "size"
    METRICS = "metrics"
    SOFTWARE_SPEC = "software_spec"
    HARDWARE_SPEC = "hardware_spec"
    PROJECT_ID = "project_id"
    SPACE_ID = "space_id"
    MODEL_DEFINITON = "model_definition"



    supported_frameworks_tar_gz = ["tensorflow","spss-modeler","pmml","caffe","caffe2",
                                   "pytorch","blueconnect","torch","mxnet","theano","darknet",
                                   "scikit-learn", "mllib"]
    @staticmethod
    def is_supported_tar_framework(framework_name):
        for each_name in MetaNames.supported_frameworks_tar_gz:
            if each_name in framework_name:
                return True
        else:
            return False

    archive_format = ["spss-modeler","pmml","caffe","caffe2","pytorch","blueconnect","torch","mxnet","theano","darknet", "mllib", "scikit-learn"]
    @staticmethod
    def is_archive_framework(framework_name):
        for each_name in MetaNames.archive_format:
            if each_name in framework_name:
                return True
        else:
            return False

    class EXPERIMENTS(object):
        TAGS = "tags"
        SETTINGS = "settings"
        TRAINING_RESULTS_REFERENCE = "trainingResultsReference"
        TRAINING_REFERENCES = "trainingReferences"
        TRAINING_DATA_REFERENCE = "trainingDataReference"
        PATCH_INPUT = "experimentPatchInput"

    class LIBRARIES(object):
        NAME = "name"
        VERSION = "version"
        DESCRIPTION = "description"
        PLATFORM = "platform"
        PATCH_INPUT = "librariesPatchInput"
        URL = "librariesUrl"
        CONTENT_URL = "contentUrl"
        MODEL_DEFINITON = "model_definiton"
        TAGS = "tags"
        SPACE_URL = "spaceUrl"
        COMMAND = "command"
        CUSTOM = "custom"

    class RUNTIMES(object):
        NAME = "name"
        DESCRIPTION = "description"
        PLATFORM = "platform"
        CUSTOM_LIBRARIES_URLS = "customLibrariesUrls"
        PATCH_INPUT = "runtimesPatchInput"
        URL = "runtimesUrl"
        CONTENT_URL = "contentUrl"

    class FUNCTIONS(object):
        REVISION = "revision"
        REVISION_URL = "revisionUrl"
        URL = "functionUrl"
        TYPE = "type"
        CONTENT_URL = "contentUrl"
        DESCRIPTION = "description"
        INPUT_DATA_SCHEMA = "inputDataSchema"
        OUTPUT_DATA_SCHEMA = "outputDataSchema"
        TAGS = "tags"
        SAMPLE_SCORING_INPUT = "sampleScoringInput"


#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import sys
import os
import time
import logging
import argparse
from zipfile import ZipFile

fl_path = os.path.abspath('.')
if fl_path not in sys.path:
    sys.path.append(fl_path)

import ibmfl.envs as fl_envs
from ibmfl._version import __version__
from ibmfl.connection.router_handler import Router
from ibmfl.message.message import Message
from ibmfl.message.message_type import MessageType
from ibmfl.exceptions import FLException, InvalidConfigurationException
from ibmfl.connection.route_declarations import get_party_router
import ibmfl.util.config as fl_config
from ibmfl.util.config import read_yaml_config, get_class_by_name, \
	configure_logging_from_file, get_party_config, get_lt_from_config, convert_bytes_to_zip

logger = logging.getLogger(__name__)


class Party:
    """
    Party participates in Federated Learning training.  
    """

    def __init__(self, **kwargs):
        """
        Initializes a party based on the configuration provided in 
        either a yaml file or a dictionary.
        :param config_file: path to yml file with the configuration of the party
        :type config_file: `str`
        :param config_dict: dictionary containing the configuration of the party
        :type config_dict: `dict`
        :return None
        """
        log_level = kwargs.get('log_level', 'INFO')
        configure_logging_from_file(log_level=log_level)

        config_file = kwargs.get('config_file')
        config_dict = kwargs.get('config_dict')
        if config_file != None:
            config_dict = fl_config.read_yaml_config(config_file=config_file)

        config_dict['connection']['name'] = 'SWSConnection'
        config_dict['connection']['path'] = 'ibmfl.connection.websockets_connection'
        cls_config = fl_config.get_cls_by_config(config_dict)

        self.party_config = cls_config

        self.data_handler = None
        self.fl_model = None
        self.is_running = False

        data_config = cls_config.get('data')
        connection_config = cls_config.get('connection')
        ph_config = cls_config.get('protocol_handler')
        mrec_config = cls_config.get('metrics_recorder')

        try:
            # Load data (optional field)
            # - in some cases the aggregator doesn't have data for testing purposes
            data_cls_ref = data_config.get('cls_ref')
            data_info = data_config.get('info')
            self.data_handler = data_cls_ref(data_config=data_info)
            # Read and create model (optional field)
            # In some cases aggregator doesn't need to load the model:

            self.fl_model = None

            # Load hyperparams
            self.hyperparams = cls_config.get('hyperparams')
            self.agg_info = cls_config.get('aggregator')
            self.agg_info["ip"] = self.agg_info["ip"] + "/aggregator"
            self.agg_info["port"] = 443

            connection_cls_ref = connection_config.get('cls_ref')
            self.connection_info = connection_config.get('info')
            self.connection_info["VERSION"] = __version__
            self.connection_info["ip"] = self.agg_info['ip']
            self.connection_info["port"] = self.agg_info['port']
            connection_synch = connection_config.get('sync')
            # Use opt-in privacy for WML, to get metrics by default
            connection_private = self.connection_info.get('private',False)
            self.connection = connection_cls_ref(self.connection_info)
            self.connection.initialize_sender()

            self.local_training_handler = None

            ph_cls_ref = ph_config.get('cls_ref')
            if mrec_config:
                mrec_cls_ref = mrec_config.get('cls_ref')
                metrics_recorder = mrec_cls_ref(mrec_config.get('output_file'),
                                                 mrec_config.get('output_type'),
                                                 mrec_config.get('compute_pre_train_eval'),
                                                 mrec_config.get('compute_post_train_eval'))
            else:
                metrics_recorder = None


            self.proto_handler = ph_cls_ref(self.fl_model,
                                            self.connection.sender,
                                            self.data_handler,
                                            self.local_training_handler,
                                            metrics_recorder,
                                            agg_info=self.agg_info,
                                            synch=connection_synch,
                                            is_private=connection_private)

            self.router = Router()
            get_party_router(self.router, self.proto_handler)

            # check for token
            token = kwargs.get('token', None)
            self_signed_cert_flag = kwargs.get('self_signed_cert', None)
            #self.connection.initialize(router=self.router, aggIinfo=self.agg_info, ssl_context=self_signed_cert_flag, authToken=token)
            self.connection.initialize_receiver(router=self.router, verify=not self_signed_cert_flag, token=token)

            self.connection.start()

        except Exception as ex:
            logger.info('Error occurred '
                        'while loading aggregator configuration')
            logger.exception(ex)
        else:
            logger.info("Party initialization successful")

        self.agg_info = cls_config.get('aggregator')

    def initialize_model_config(self):
        logger.info('Initializing model configuration')
        cls_config = self.party_config
        model_config = cls_config.get('model')
        lt_config = cls_config.get('local_training')
        try:
            # Read and create model (optional field)
            # In some cases aggregator doesn't need to load the model:
            model_cls_ref = model_config.get('cls_ref')
            spec = model_config.get('spec')
            self.fl_model = model_cls_ref('', spec)

        except Exception as ex:
            logger.error("Error occurred while loading model config. Please check for missing libraries.")
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading model config.')

        try:
            self.hyperparams = cls_config.get('hyperparams')

            lt_cls_ref = lt_config.get('cls_ref')
            self.local_training_handler = lt_cls_ref(
                self.fl_model, self.data_handler)

            self.proto_handler.set_model(self.fl_model)
            self.proto_handler.set_training_handler(self.local_training_handler)

        except Exception as ex:
            logger.info('Error occurred '
                        'while loading model and lth configuration')
            logger.exception(ex)
        else:
            logger.info("Party model initialization successful")

    def register_party(self):
        """
        Registers party with the aggregator.

        :param: None
        :return: None
        """
        logger.info('Registering party...')
        returnValue = False
        register_message = Message(
            MessageType.REGISTER.value, data=self.connection_info)
        try:
            response = self.connection.sender.send_message(
                self.agg_info, register_message)
            if response is not None and response.get_data()['status'] == 'success':

                data = response.get_data()
                if self.extract_config_from_stream(data):
                    logger.info('Registration Successful')
                    returnValue = True
                else:
                    logger.info('Registration Failed: Model processing error')
                    returnValue = False
            else:
                logger.error('Registration Failed: Failure status from aggregator')
                if response is not None:
                    logger.error(response.get_data().get('detail', "No Detail Provided"))
        except Exception as ex:
            logger.error("Error occurred during registration" + str(ex))

        return returnValue

    def stop_connection(self):
        """
        Stop the connection to the aggregator

        :param: None
        :return: None
        """
        try:
            self.connection.stop()

        except Exception as ex:
            logger.error("Error occurred during stop")
            logger.error(ex)
        else:
            logger.info("Party stop successful")

    def evaluate_model(self):
        """
        Calls function that evaluates current model with local testing data
        and prints the results.

        :param: None
        :return: None
        """
        try:
            self.proto_handler.print_evaluate_local_model()

        except Exception as ex:
            logger.error("Error occurred during evaluation.")
            logger.error(ex)

    def start(self):
        """
        Initializes connection and registers with the aggregator,
        then accept commands from the aggregator to effect training.  

        :param: None
        :return: None
        """

        if not self.is_running:
            logger.info('Party not registered yet.')
            if self.register_party():
                logger.info('Listening for commands from Aggregator')
                self.is_running = True
                self.initialize_model_config()
            else:
                self.stop_connection()
                raise FLException(
                    "Error occurred while registering party. ")
        else:
            logger.info('Party already running.')

    def extract_config_from_stream(self, data):
        """Read model from response stream and extract the content.
        :param data: response data recieved from aggregator.
        :type data: `json` 
        :return: None
        """
        returnValue = False
        working_dir = fl_envs.working_directory

        if 'model_package' in data:
            logger.info(
                'Model Package found, placing in working_dir ' + working_dir)
            model_file_name = 'model_package.zip'
            model_file_op = os.path.join(working_dir, model_file_name)
            model_pk_bytes = data.get('model_package')

            try:
                convert_bytes_to_zip(model_pk_bytes, model_file_op)

                with ZipFile(model_file_op, 'r') as o:
                    o.extractall(working_dir)

            except Exception as ex:
                logger.exception(
                    "Error occurred while unpacking model package from aggregator.")
                raise FLException(
                    "Error occurred while downloading model package. ")
        else:
            logger.info("No model package received from aggregator.")

        logger.info('Checking for model config in payload.')

        model_config_agg = data.get('model_config', None)
        if model_config_agg == None:
            logger.info("No model config received from aggregator.")
            return returnValue

        if 'is_model_working_dir' in model_config_agg.get('spec'):
            model_config_agg['spec']['model_definition'] = working_dir

        elif 'model_definition' in  model_config_agg.get('spec'):
            model_def = os.path.join(working_dir, model_config_agg['spec']['model_definition'])
            model_config_agg['spec']['model_definition'] = model_def

        logger.info(model_config_agg)
        self.party_config['model'] = model_config_agg

        local_training_agg = data.get('local_training')
        logger.info("Default local training: %s" % str(local_training_agg))

        if self.party_config.get('local_training'):
            logger.info("Using configured local training: %s" % str(self.party_config.get('local_training')))
        else:
            logger.info("Using default local training from aggregator")
            if local_training_agg == None:
                logger.error("No local training config received from aggregator.")
                return returnValue
            self.party_config['local_training'] = get_lt_from_config(local_training_agg)

        logger.info('config extract finished!')

        returnValue = True
        return returnValue


if __name__ == '__main__':
    """
    Main function to run a party-side application in a Federated Learning training job.
    The application can either be run in an interactive ( -i ), or a non-interactive
    mode.  

    In interactive mode, enter commands:

    START     - Start the connection for communication with the aggregator
    STOP      - Stop the connection 
    REGISTER  - Register with the aggregator and start accepting commands
    EVAL      - Evaluate the model with local test data

    In non-interactive mode, the party will be started, and registered with
    the aggregator.  After training is complete and a STOP message is received
    from the aggregator, the application will exit.

    """

    parser = argparse.ArgumentParser(
        description='WML Federated Learning Party')
    parser.add_argument('config_file', help='yaml configuration file')
    parser.add_argument('token', help='authentication token')
    parser.add_argument('-s', '--self_signed_cert', help='flag for self-signed certificate', action='store_true')
    parser.add_argument('-i', '--interactive',
                        help='run interactively', action='store_true')
    parser.add_argument('-l', '--log_level', type=str, default="INFO", required=False,
                        help='log_level should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]')
    args = parser.parse_args()

    if (args.self_signed_cert):
        self_signed_cert_arg = True
    else:
        self_signed_cert_arg = None

    p = Party(config_file=args.config_file, token=args.token, self_signed_cert=self_signed_cert_arg,
              log_level=args.log_level)

    if (args.interactive):
        # Indefinite loop to accept user commands to execute
        while 1:

            try:
                msg = sys.stdin.readline()
                if re.match('START', msg):
                    p.start()

                if re.match('STOP', msg):
                    p.stop_connection()
                    break

                if re.match('REGISTER', msg):
                    p.register_party()

                if re.match('EVAL', msg):
                    p.evaluate_model()

            except Exception as ex:
                logger.error("Error occurred during " + msg)
                logger.error(ex)
    else:
        try:
            p.start()
        except Exception as ex:
            logger.error("Error occurred during start")
            logger.error(ex)
            p.stop_connection()

        while not p.connection.stopped:
            logger.debug("Party is still running")
            time.sleep(10)
        logger.info("Party is complete")



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, save_data_to_container
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestWebservice, \
    AbstractTestAutoAIAsync

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms


class TestAutoAIRemote(AbstractTestAutoAIAsync, AbstractTestWebservice, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/arabicghosts_train.csv'

    cos_endpoint = "https://s3.us.cloud-object-storage.appdomain.cloud"
    data_cos_path = 'data/arabicghosts_train.csv'

    SPACE_ONLY = False

    OPTIMIZER_NAME = "Arabic Ghost test sdk"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.MULTICLASS,
        prediction_column='لون',
        scoring=Metrics.LOG_LOSS,
        holdout_size=0.2,
        include_only_estimators=[ClassificationAlgorithms.SnapRF, ClassificationAlgorithms.SnapDT,
                                 ClassificationAlgorithms.SnapSVM, ClassificationAlgorithms.SnapLR],
        max_number_of_estimators=4
    )

    def test_00b_write_data_to_container(self):
        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

        save_data_to_container(self.data_location, self.data_cos_path, self.wml_client)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            location=ContainerLocation(path=self.data_cos_path
                                       ))
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, Union, List, Optional

from pandas import DataFrame

from .base_deployment import BaseDeployment
from ..wml_client_error import WMLClientError

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline
    from pandas import DataFrame
    from numpy import ndarray
    from ..workspace import WorkSpace

__all__ = [
    "WebService"
]


class WebService(BaseDeployment):
    """
    An Online Deployment class aka. WebService.
    With this class object you can manage any online (WebService) deployment.

    Parameters
    ----------
    source_wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance where training was performed.

    source_project_id: str, optional
        ID of the Watson Studio project where training was performed.

    source_space_id: str, optional
        ID of the Watson Studio Space where training was performed.

    target_wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance where you want to deploy.

    target_project_id: str, optional
        ID of the Watson Studio project where you want to deploy.

    target_space_id: str, optional
        ID of the Watson Studio Space where you want to deploy.
    """

    def __init__(self,
                 source_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 source_project_id: str = None,
                 source_space_id: str = None,
                 target_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 target_project_id: str = None,
                 target_space_id: str = None,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None):

        super().__init__(
            deployment_type='online',
            source_wml_credentials=source_wml_credentials,
            source_project_id=source_project_id,
            source_space_id=source_space_id,
            target_wml_credentials=target_wml_credentials,
            target_project_id=target_project_id,
            target_space_id=target_space_id,
            wml_credentials=wml_credentials,
            project_id=project_id,
            space_id=space_id
        )

        self.name = None
        self.scoring_url = None
        self.id = None
        self.asset_id = None

    def __repr__(self):
        return f"name: {self.name}, id: {self.id}, scoring_url: {self.scoring_url}, asset_id: {self.asset_id}"

    def __str__(self):
        return f"name: {self.name}, id: {self.id}, scoring_url: {self.scoring_url}, asset_id: {self.asset_id}"

    def create(self,
               model: str,
               deployment_name: str,
               serving_name: str = None,
               metadata: Optional[Dict] = None,
               training_data: Optional[Union['DataFrame', 'ndarray']] = None,
               training_target: Optional[Union['DataFrame', 'ndarray']] = None,
               experiment_run_id: Optional[str] = None) -> None:
        """
        Create deployment from a model.

        Parameters
        ----------
        model: str, required
            AutoAI model name.

        deployment_name: str, required
            Name of the deployment

        training_data: Union['pandas.DataFrame', 'numpy.ndarray'], optional
            Training data for the model

        training_target: Union['pandas.DataFrame', 'numpy.ndarray'], optional
            Target/label data for the model

        serving_name: str, optional
            Serving name of the deployment

        metadata: dictionary, optional
            Model meta properties.

        experiment_run_id: str, optional
            ID of a training/experiment (only applicable for AutoAI deployments)

        Example
        -------
        >>> from ibm_watson_machine_learning.deployment import WebService
        >>>
        >>> deployment = WebService(
        >>>        wml_credentials={
        >>>              "apikey": "...",
        >>>              "iam_apikey_description": "...",
        >>>              "iam_apikey_name": "...",
        >>>              "iam_role_crn": "...",
        >>>              "iam_serviceid_crn": "...",
        >>>              "instance_id": "...",
        >>>              "url": "https://us-south.ml.cloud.ibm.com"
        >>>            },
        >>>         project_id="...",
        >>>         space_id="...")
        >>>
        >>> deployment.create(
        >>>        experiment_run_id="...",
        >>>        model=model,
        >>>        deployment_name='My new deployment'
        >>>    )
        """
        return super().create(model=model,
                              deployment_name=deployment_name,
                              metadata=metadata,
                              serving_name=serving_name,
                              training_data=training_data,
                              training_target=training_target,
                              experiment_run_id=experiment_run_id,
                              deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def get_params(self) -> Dict:
        """Get deployment parameters."""
        return super().get_params()

    @BaseDeployment._project_to_space_to_project
    def score(self, payload: 'DataFrame', transaction_id: str = None) -> Dict[str, List]:
        """
        Online scoring on WML. Payload is passed to the WML scoring endpoint where model have been deployed.

        Parameters
        ----------
        payload: pandas.DataFrame, required
            DataFrame with data to test the model.

        transaction_id: str, optional
            Can be used to indicate under which id the records will be saved into payload table in IBM OpenScale.

        Returns
        -------
        Dictionary with list od model output/predicted targets.

        Example
        -------
        >>> predictions = deployment.score(payload=test_data)
        >>> print(predictions)
            {'predictions':
                [{
                    'fields': ['prediction', 'probability'],
                    'values': [['no', [0.9221385608558003, 0.07786143914419975]],
                              ['no', [0.9798324002736079, 0.020167599726392187]]
                }]}
        """
        return super().score(payload=payload, transaction_id=transaction_id)

    @BaseDeployment._project_to_space_to_project
    def delete(self, deployment_id: str = None) -> None:
        """
        Delete deployment on WML.

        Parameters
        ----------
        deployment_id: str, optional
            ID of the deployment to delete. If empty, current deployment will be deleted.

        Example
        -------
        >>> deployment = WebService(workspace=...)
        >>> # Delete current deployment
        >>> deployment.delete()
        >>> # Or delete a specific deployment
        >>> deployment.delete(deployment_id='...')
        """
        super().delete(deployment_id=deployment_id, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def list(self, limit=None) -> 'DataFrame':
        """
        List WML deployments.

        Parameters
        ----------
        limit: int, optional
            Set the limit of how many deployments to list. Default is None (all deployments should be fetched)

        Returns
        -------
        Pandas DataFrame with information about deployments.

        Example
        -------
        >>> deployment = WebService(workspace=...)
        >>> deployments_list = deployment.list()
        >>> print(deployments_list)
                             created_at  ...  status
            0  2020-03-06T10:50:49.401Z  ...   ready
            1  2020-03-06T13:16:09.789Z  ...   ready
            4  2020-03-11T14:46:36.035Z  ...  failed
            3  2020-03-11T14:49:55.052Z  ...  failed
            2  2020-03-11T15:13:53.708Z  ...   ready
        """
        return super().list(limit=limit, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def get(self, deployment_id: str) -> None:
        """
        Get WML deployment.


        Parameters
        ----------
        deployment_id: str, required
            ID of the deployment to work with.

        Returns
        -------
        WebService deployment object

        Example
        -------
        >>> deployment = WebService(workspace=...)
        >>> deployment.get(deployment_id="...")
        """
        super().get(deployment_id=deployment_id, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def _deploy(self,
                pipeline_model: 'Pipeline',
                deployment_name: str,
                meta_props: Dict,
                serving_name=None,
                result_client=None) -> Dict:
        """
        Deploy model into WML.

        Parameters
        ----------
        pipeline_model: Union['Pipeline', str], required
            Model of the pipeline to deploy

        deployment_name: str, required
            Name of the deployment

        meta_props: dictionary, required
            Model meta properties.

        serving_name: str, required
            Serving name of the deployment

        result_client: Tuple['DataConnection', 'resource'] required
            Tuple with Result DataConnection object and initialized COS client.
        """
        deployment_details = {}
        asset_uid = self._publish_model(pipeline_model=pipeline_model,
                                        meta_props=meta_props)

        self.asset_id = asset_uid

        conf_names = self._target_workspace.wml_client.deployments.ConfigurationMetaNames

        deployment_props = {
            conf_names.NAME: deployment_name,
            conf_names.ONLINE: {},
            "hardware_spec": {
                "name": "M",
            }
        }

        if serving_name:
            deployment_props[conf_names.ONLINE]["parameters"] = {conf_names.SERVING_NAME: serving_name}

        print("Deploying model {} using V4 client.".format(asset_uid))
        try:

            deployment_details = self._target_workspace.wml_client.deployments.create(
                artifact_uid=asset_uid,
                meta_props=deployment_props)
            self.deployment_id = self._target_workspace.wml_client.deployments.get_uid(deployment_details)

        except WMLClientError as e:
            raise e

        return deployment_details



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import logging
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *

class TestExportImport(unittest.TestCase):
    model_id1 = None
    model_id2 = None

    export_space_id = None
    import_space_id = None

    model_id_for_deployment = None
    deployment_id = None
    scoring_url = None

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestExportImport.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.cos_credentials = get_cos_credentials()
        self.cos_resource_crn = self.cos_credentials['resource_instance_id']
        self.instance_crn = get_instance_crn()

        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'space_export_' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description',
                     self.client.spaces.ConfigurationMetaNames.STORAGE: {
                                                                           "resource_crn": self.cos_resource_crn
                                                                        },
                     self.client.spaces.ConfigurationMetaNames.COMPUTE: {
                         "name": "existing_instance_id",
                         "crn": self.instance_crn
            }
        }

        self.space = self.client.spaces.store(meta_props=metadata, background_mode=False)

        TestExportImport.export_space_id = self.client.spaces.get_id(self.space)
        print("export space_id: ", TestExportImport.export_space_id)

        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'space_import_' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description',
                     self.client.spaces.ConfigurationMetaNames.STORAGE: {
                                                                           "resource_crn": self.cos_resource_crn
                                                                        },
                     self.client.spaces.ConfigurationMetaNames.COMPUTE: {
                         "name": "existing_instance_id",
                         "crn": self.instance_crn
            }
        }

        self.space = self.client.spaces.store(meta_props=metadata, background_mode=False)

        TestExportImport.import_space_id = self.client.spaces.get_id(self.space)
        print("import space_id: ", TestExportImport.import_space_id)

    def test_01_publish_local_model_in_repository(self):

        self.client.set.default_space(TestExportImport.export_space_id)

        self.client.repository.ModelMetaNames.show()

        # model_content_path = 'artifacts/mnistCNN.h5.tgz'
        model_content_path = 'artifacts/tf_model_fvt_test.tar.gz'

        base_sw_spec_id = self.client.software_specifications.get_uid_by_name('default_py3.7')
        # base_sw_spec_id = self.client.software_specifications.get_id_by_name("ai-function_0.1-py3.6")

        print(base_sw_spec_id)

        meta_prop_sw_spec = {
            self.client.software_specifications.ConfigurationMetaNames.NAME: "test_sw_spec_" + str(uuid.uuid4()),
            self.client.software_specifications.ConfigurationMetaNames.DESCRIPTION: "Software specification for test",
            self.client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION: {"guid": base_sw_spec_id}
        }

        sw_spec_details = self.client.software_specifications.store(meta_props=meta_prop_sw_spec)

        print(sw_spec_details)

        TestExportImport.sw_spec_id = self.client.software_specifications.get_id(sw_spec_details)

        model_details = self.client.repository.store_model(model=model_content_path,
                                                           meta_props={self.client.repository.ModelMetaNames.NAME: "test_keras1",
                                                                  # self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: '2b73a275-7cbf-420b-a912-eae7f436e0bc',
                                                                       self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: TestExportImport.sw_spec_id,
                                                                       self.client.repository.ModelMetaNames.TYPE: 'tensorflow_2.1'})
        # self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: TestExportImport.sw_spec_id,
        #                                                           self.client.repository.ModelMetaNames.TYPE: 'keras_2.2.5'})
        print("model_id1: ", model_details)
        TestExportImport.model_id1 = self.client.repository.get_model_uid(model_details)

        print(TestExportImport.model_id1)

        self.assertIsNotNone(TestExportImport.model_id1)

        model_details = self.client.repository.store_model(model=model_content_path,
                                                           meta_props={self.client.repository.ModelMetaNames.NAME: "test_keras2",
                                                                  # self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: '2b73a275-7cbf-420b-a912-eae7f436e0bc',
                                                                       self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: TestExportImport.sw_spec_id,
                                                                       self.client.repository.ModelMetaNames.TYPE: 'tensorflow_2.1'})
        # self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: TestExportImport.sw_spec_id,
        #                                                           self.client.repository.ModelMetaNames.TYPE: 'keras_2.2.5'})
        print("model_id2: ", model_details)
        TestExportImport.model_id2 = self.client.repository.get_model_uid(model_details)

        print(TestExportImport.model_id2)

        self.assertIsNotNone(TestExportImport.model_id2)

    def test_02_deployment_of_model(self):

        deployment_details = self.client.deployments.create(TestExportImport.model_id1,
                                                            meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test keras to be exported deployment",
                                                                        self.client.deployments.ConfigurationMetaNames.ONLINE:{}})

        print(deployment_details)

        TestExportImport.deployment_id = self.client.deployments.get_id(deployment_details)
        TestExportImport.scoring_url = self.client.deployments.get_scoring_href(deployment_details)

        self.assertTrue('online' in str(deployment_details))

        deployment_details = self.client.deployments.get_details(deployment_uid=TestExportImport.deployment_id)

        print(deployment_details)

        self.client.deployments.delete(TestExportImport.deployment_id)


    def test_03_export(self):
        metadata = { self.client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                     # self.client.export_assets.ConfigurationMetaNames.ASSET_TYPES: ["wml_function"],
                     self.client.export_assets.ConfigurationMetaNames.ASSET_IDS: [TestExportImport.model_id1,
                                                                                  TestExportImport.model_id2]
                     # self.client.export_assets.ConfigurationMetaNames.ASSET_TYPES: ["wml_model"]
         # self.client.export_assets.ConfigurationMetaNames.ALL_ASSETS: True
                     }
        # self.client.export_assets.ConfigurationMetaNames.ASSET_TYPES: ["wml_model"],
        # self.client.export_assets.ConfigurationMetaNames.ASSET_IDS: ["13a53931-a8c0-4c2f-8319-c793155e7517",
        #                                                             "13a53931-a8c0-4c2f-8319-c793155e7518"]
        #                                                            >> >}
        details = self.client.export_assets.start(meta_props=metadata, space_id=TestExportImport.export_space_id)
        print(details)

        TestExportImport.export_job_id = details[u'metadata'][u'id']

        import time

        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 10 * 60:
            time.sleep(3)
            response = self.client.export_assets.get_details(TestExportImport.export_job_id, space_id=TestExportImport.export_space_id)
            state = response[u'entity'][u'status'][u'state']
            print(state)
            if state == 'completed' or state == 'error' or state == 'failed':
                break
            diff_time = time.time() - start_time

        print(response)

        self.assertTrue(state == 'completed')

        details = self.client.export_assets.get_exported_content(TestExportImport.export_job_id,
             space_id = TestExportImport.export_space_id,
             file_path = '/Users/mitun_vb/export_import/data/my_exported_content.zip')

        print(details)

    def test_04_import(self):
        self.client.set.default_space(TestExportImport.import_space_id)
        self.client.repository.list_models()
        self.client.software_specifications.list()

        details = self.client.import_assets.start(file_path='/Users/mitun_vb/export_import/data/my_exported_content.zip',
                                                  space_id=TestExportImport.import_space_id)
        print(details)

        TestExportImport.import_job_id = details[u'metadata'][u'id']

        import time

        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 10 * 60:
            time.sleep(3)
            response = self.client.import_assets.get_details(TestExportImport.import_job_id,
                                                             space_id=TestExportImport.import_space_id)
            state = response[u'entity'][u'status'][u'state']
            print(state)
            if state == 'completed' or state == 'error' or state == 'failed':
                break
            diff_time = time.time() - start_time

        print(response)

        self.client.repository.list_models()
        self.client.software_specifications.list()
        details = self.client.repository.get_model_details()
        print(details)


        for obj in details[u'resources']:
            if obj[u'metadata'][u'name'] == "test_keras1":
                TestExportImport.model_id_for_deployment = obj[u'metadata'][u'id']

        print(TestExportImport.model_id_for_deployment)

    def test_05_deployment_of_imported_model(self):

        deployment_details = self.client.deployments.create(TestExportImport.model_id_for_deployment,
                                                            meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test keras imported deployment",
                                                                        self.client.deployments.ConfigurationMetaNames.ONLINE:{}})

        print(deployment_details)

        TestExportImport.deployment_id = self.client.deployments.get_id(deployment_details)
        TestExportImport.scoring_url = self.client.deployments.get_scoring_href(deployment_details)

        self.assertTrue('online' in str(deployment_details))

        deployment_details = self.client.deployments.get_details(deployment_uid=TestExportImport.deployment_id)

        print(deployment_details)

        self.client.deployments.delete(TestExportImport.deployment_id)

    # def test_05_delete_model(self):
    #     TestExportImport.logger.info("Delete model")
    #     self.client.repository.delete(TestExportImport.model_id1)
    #     self.client.repository.delete(TestExportImport.model_id2)

    def test_06_delete_space(self):
        TestExportImport.logger.info("Delete space and jobs")
        self.client.export_assets.list(space_id=TestExportImport.export_space_id)
        self.client.import_assets.list(space_id=TestExportImport.import_space_id)

        self.client.export_assets.delete(TestExportImport.export_job_id, space_id=TestExportImport.export_space_id)
        self.client.import_assets.delete(TestExportImport.import_job_id, space_id=TestExportImport.import_space_id)

        self.client.spaces.delete(TestExportImport.export_space_id)
        self.client.spaces.delete(TestExportImport.import_space_id)

if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import pandas as pd

from ibm_watson_machine_learning.tests.utils import get_wml_credentials, get_cos_credentials
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.experiment.autoai.engines import WMLEngine
from ibm_watson_machine_learning.helpers import S3Connection, S3Location, DataConnection
from ibm_watson_machine_learning.utils.autoai.errors import HoldoutSplitNotSupported
from ibm_watson_machine_learning.experiment.autoai.runs import AutoPipelinesRuns


class TestDataConnection(unittest.TestCase):
    data_location = './autoai/data/bank.csv'
    data = None
    bank_connection: 'DataConnection' = None
    wml_credentials_cp4d = None
    csv_with_different_delimeter_location = './autoai/data/datasets_different_delimeter/iris_dataset_different_delimeter.csv'

    @classmethod
    def setUp(cls) -> None:
        cls.data = pd.read_csv(cls.data_location)
        cls.X = cls.data.drop(['y'], axis=1)
        cls.y = cls.data['y']
        cls.wml_credentials = get_wml_credentials('CLOUD_PROD_AM')
        cls.wml_credentials_cp4d = get_wml_credentials('CLOUD_DEV_AM')
        cls.cos_credentials = get_cos_credentials('CLOUD_PROD_AM')

    def test_01_initialize_DataConnection__object_initialized(self):
        TestDataConnection.bank_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='bank.csv')
        )
        self.assertEqual(self.bank_connection.connection.endpoint_url, self.cos_credentials['endpoint_url'],
                         msg="Wrong endpoint_url stored.")
        self.assertEqual(self.bank_connection.connection.access_key_id, self.cos_credentials['access_key_id'],
                         msg="Wrong access_key_id stored.")
        self.assertEqual(self.bank_connection.connection.secret_access_key, self.cos_credentials['secret_access_key'],
                         msg="Wrong secret_access_key stored.")
        self.assertEqual(self.bank_connection.location.bucket, 'automlnew', msg="Wrong bucket stored.")
        self.assertEqual(self.bank_connection.location.path, 'bank.csv', msg="Wrong path stored.")

        print(self.bank_connection)

    def test_02_read__download_dataset_and_load_from_COS(self):
        data = self.bank_connection.read()
        print(data)

    def test_03_read_logs__download_logs_from_COS(self):
        wml_client = APIClient(self.wml_credentials)
        historical_opt = AutoPipelinesRuns(WMLEngine(wml_client=wml_client))
        data_connection = historical_opt.get_data_connections(run_id='b1308028-8bfd-44af-88d8-693697fea4bb')[0]
        data = data_connection.read_logs()
        print(data)

        self.assertIsInstance(data, str)

    def test_04_read_logs__download_logs_from_filesystem(self):
        wml_client = APIClient(self.wml_credentials_cp4d,
                                                    project_id='dd829201-9d59-4f5a-b0e0-6ea3a88ae66b')
        historical_opt = AutoPipelinesRuns(WMLEngine(wml_client=wml_client))
        data_connection = historical_opt.get_data_connections(run_id='165ff73b-77cb-49b4-aee4-4d4b87dcad37')[0]
        data = data_connection.read_logs()
        print(data)

        self.assertIsInstance(data, str)

    def test_05_read__download_dataset_stored_on_COS_with_holdout_split_recreation(self):
        auto_pipelines_params = {
            'name': 'test name',
            'desc': 'test description',
            'prediction_type': 'classification',
            'prediction_column': 'y',
            'scoring': 'roc_auc',
            'test_size': 0.1,
            'max_num_daub_ensembles': 1
        }
        self.bank_connection.auto_pipeline_params = auto_pipelines_params
        train_data, holdout_data = self.bank_connection.read(with_holdout_split=True)
        print(holdout_data)

    def test_06__write__upload_dataset_to_COS(self):
        self.bank_connection.write(data=self.data_location, remote_name='bank_test.csv')
        bank_test_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='bank_test.csv')
        )
        test_data = bank_test_connection.read()
        print(test_data)

    def test_07__write__upload_dataset_to_COS_from_pandas_DataFrame(self):
        self.bank_connection.write(data=self.data, remote_name='test_data.csv')

        bank_test_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='test_data.csv')
        )
        test_data = bank_test_connection.read()
        print(test_data)

    def test_08_read__download_xlsx_dataset_stored_on_COS(self):
        test_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='test_data.xlsx')
        )

        train_data = test_connection.read()
        print(train_data)

    def test_09_read__download_xlsx_dataset_stored_on_COS_with_holdout_split(self):
        test_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='test_data.xlsx')
        )
        auto_pipelines_params = {
            'name': 'test name',
            'desc': 'test description',
            'prediction_type': 'classification',
            'prediction_column': 'y',
            'scoring': 'roc_auc',
            'test_size': 0.1,
            'max_num_daub_ensembles': 1
        }
        test_connection.auto_pipeline_params = auto_pipelines_params

        with self.assertRaises(HoldoutSplitNotSupported):
            train_data = test_connection.read(with_holdout_split=True)

    def test_10__read__download_csv_file_with_not_standard_delimeter__return_as_pandas_DataFrame(self):
        test_connection = DataConnection(
            connection=S3Connection(endpoint_url=self.cos_credentials['endpoint_url'],
                                    access_key_id=self.cos_credentials['access_key_id'],
                                    secret_access_key=self.cos_credentials['secret_access_key']),
            location=S3Location(bucket='automlnew',
                                path='iris_dataset_different_delimeter.csv')
        )

        test_connection.write(data=self.csv_with_different_delimeter_location,
                              remote_name='iris_dataset_different_delimeter.csv')

        test_data = test_connection.read()
        print(test_data)

        self.assertGreater(len(test_data.columns), 2,
                           msg="Wrong columns number, data separator probably is wrongly interpreted.")


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest,time

import logging
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *

class TestScripts(unittest.TestCase):
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestScripts.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.project_id = get_project_id()

        self.client.set.default_project(self.project_id)

        # print(self.client.service_instance.get_details())

        print("project_id: ", self.project_id)

    def test_01_create_script_asset(self):

        self.client.script.ConfigurationMetaNames.show()
        sw_spec_uid = self.client.software_specifications.get_uid_by_name("ai-function_0.1-py3.6")

        print("sw_spec_uid: ", sw_spec_uid)

        meta_prop_script = {
            self.client.script.ConfigurationMetaNames.NAME: "my script asset",
            self.client.script.ConfigurationMetaNames.DESCRIPTION: "script asset for deployment",
            self.client.script.ConfigurationMetaNames.SOFTWARE_SPEC_UID: sw_spec_uid
        }

        script_details = self.client.script.store(meta_prop_script, file_path="artifacts/test.py")

        TestScripts.script_asset_id = self.client.script.get_id(script_details)
        TestScripts.script_asset_url = self.client.script.get_href(script_details)
        TestScripts.logger.info("script asset ID:" + str(TestScripts.script_asset_id))
        TestScripts.logger.info("script asset URL:" + str(TestScripts.script_asset_url))
        self.assertIsNotNone(TestScripts.script_asset_id)
        self.assertIsNotNone(TestScripts.script_asset_url)


    def test_02_download_script_content(self):
        try:
            os.remove('test_script_asset.zip')
        except:
            pass
        self.client.script.download(TestScripts.script_asset_id, filename='test_script_asset.zip')
        # try:
        #     os.remove('test_script_asset.zip')
        # except:
        #     pass

    def test_04_get_details(self):

        details = self.client.script.get_details(TestScripts.script_asset_id)
        print(details)
        self.assertTrue(TestScripts.script_asset_id in str(details))

    def test_05_list(self):
        self.client.script.list()

    def test_06_revisions(self):
        meta_prop_script1_revision1 = {
            self.client.shiny.ConfigurationMetaNames.NAME: "Script revision 1",
            self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "Script revision 1"
        }

        # Update meta and attachment for revision 1 creation
        script_update = self.client.script.update(TestScripts.script_asset_id,
                                                  meta_prop_script1_revision1,
                                                  file_path="artifacts/test1.py.zip")
        attachment_id = script_update[u'metadata'][u'attachment_id']

        # Create revision
        revision = self.client.script.create_revision(TestScripts.script_asset_id)

        self.assertTrue(revision[u'metadata'][u'revision_id'] and revision[u'metadata'][u'revision_id'] == 1)

        new_attachment_id = revision[u'metadata'][u'attachment_id']

        self.assertTrue(attachment_id != new_attachment_id)

        # List revisions
        self.client.script.list_revisions(TestScripts.script_asset_id)

        try:
            os.remove('script1.zip')
        except:
            pass

        # Download revision 1 attachment
        self.client.script.download(TestScripts.script_asset_id,
                                    filename='script1.zip',
                                    rev_uid=1)
        try:
            os.remove('script1.zip')
        except:
            pass

    def test_15_delete_script_asset(self):
        TestScripts.logger.info("Delete function")
        self.client.script.delete(TestScripts.script_asset_id)

if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
import argparse
import re
import pip

def get_required_packages(filename):

    def parse_package_name(package_name):
        package_name_split = re.split('>=|<=', package_name)
        package_name_clean = package_name_split[0]
        if len(package_name_split) > 1:
            print('Version restrictions "{}" found in {} are not supported'.format(', '.join(package_name_split[1:]), package_name_clean))
        if '==' in package_name_clean:
            pkg = dict(zip(['name', 'version'], package_name_clean.split('==')))
        else:
            pkg = {'name': package_name_clean, 'version': None}
        return pkg

    try:
        with open(filename) as f:
            packages_from_requirements = f.read().splitlines()
    except IOError:
        print('Could not read file: {}'.format(filename))

    return [parse_package_name(pkg) for pkg in packages_from_requirements]

def get_installed_packages():
    return [{'name': pkg.project_name, 'version': pkg.version} for pkg in pip.get_installed_distributions()]

def get_missing_packages(installed_pkgs, required_pkgs):

    def infer_action(pkg):
        installed_pkg = None
        for p in installed_pkgs:
            if p['name'] == pkg['name']:
                installed_pkg = p

        if installed_pkg is None:
            return {'action': 'install'}
        else:
            if (pkg['version'] is not None) and (pkg['version'] != installed_pkg['version']):
                    return {'action': 'reinstall'}
        return {'action': None}

    missing_pkgs = []
    for pkg in required_pkgs:
        pkg_action = infer_action(pkg)
        if pkg_action['action'] is not None:
            missing_pkgs.append(dict(pkg, **pkg_action))

    return missing_pkgs

def run(filename=None):

    required_packages = get_required_packages(filename)
    installed_packages = get_installed_packages()
    missing_packages = get_missing_packages(installed_packages, required_packages)

    if len(missing_packages) > 0:
        if 'win' in sys.platform:
            print('Automatic installation is not supported on Windows')
            print('Check if following packages are installed: \n{}'.format('\n'.join(packages_from_requirements)))
            quit()

        for pkg in missing_packages:
            if pkg['action'] == 'reinstall':
                pip_uninstall_arg = ['uninstall']
                pip_uninstall_arg.append('{}'.format(pkg['name']))
                pip_uninstall_arg.append('-y')
                pip.main(pip_uninstall_arg)
            pip_install_arg = ['install']
            if pkg['version'] is None:
                pip_install_arg.append('{}'.format(pkg['name']))
            else:
                pip_install_arg.append('{}=={}'.format(pkg['name'], pkg['version']))
            pip.main(pip_install_arg)
    else:
        print("All required packages are installed.")

if __name__ == '__main__':

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-r', type=str, default='requirements.txt', help='Requirements file name')
    FLAGS, unparsed = arg_parser.parse_known_args()
    run(filename=FLAGS.r)



# coding: utf-8

"""

    No descripton provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems
import re


class ExperimentStatusExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, submitted_at=None, finished_at=None, current_at=None, current_iteration=None, total_iterations=None, state=None, message=None, failure=None):
        """
        ExperimentStatusExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'submitted_at': 'datetime',
            'finished_at': 'datetime',
            'current_at': 'datetime',
            'current_iteration': 'float',
            'total_iterations': 'float',
            'state': 'str',
            'message': 'str',
            'failure': 'ErrorSchemaExperiments'
        }

        self.attribute_map = {
            'submitted_at': 'submitted_at',
            'finished_at': 'finished_at',
            'current_at': 'current_at',
            'current_iteration': 'current_iteration',
            'total_iterations': 'total_iterations',
            'state': 'state',
            'message': 'message',
            'failure': 'failure'
        }

        self._submitted_at = submitted_at
        self._finished_at = finished_at
        self._current_at = current_at
        self._current_iteration = current_iteration
        self._total_iterations = total_iterations
        self._state = state
        self._message = message
        self._failure = failure

    @property
    def submitted_at(self):
        """
        Gets the submitted_at of this ExperimentStatusExperiments.


        :return: The submitted_at of this ExperimentStatusExperiments.
        :rtype: datetime
        """
        return self._submitted_at

    @submitted_at.setter
    def submitted_at(self, submitted_at):
        """
        Sets the submitted_at of this ExperimentStatusExperiments.


        :param submitted_at: The submitted_at of this ExperimentStatusExperiments.
        :type: datetime
        """

        self._submitted_at = submitted_at

    @property
    def finished_at(self):
        """
        Gets the finished_at of this ExperimentStatusExperiments.


        :return: The finished_at of this ExperimentStatusExperiments.
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """
        Sets the finished_at of this ExperimentStatusExperiments.


        :param finished_at: The finished_at of this ExperimentStatusExperiments.
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def current_at(self):
        """
        Gets the current_at of this ExperimentStatusExperiments.


        :return: The current_at of this ExperimentStatusExperiments.
        :rtype: datetime
        """
        return self._current_at

    @current_at.setter
    def current_at(self, current_at):
        """
        Sets the current_at of this ExperimentStatusExperiments.


        :param current_at: The current_at of this ExperimentStatusExperiments.
        :type: datetime
        """

        self._current_at = current_at

    @property
    def current_iteration(self):
        """
        Gets the current_iteration of this ExperimentStatusExperiments.


        :return: The current_iteration of this ExperimentStatusExperiments.
        :rtype: float
        """
        return self._current_iteration

    @current_iteration.setter
    def current_iteration(self, current_iteration):
        """
        Sets the current_iteration of this ExperimentStatusExperiments.


        :param current_iteration: The current_iteration of this ExperimentStatusExperiments.
        :type: float
        """

        self._current_iteration = current_iteration

    @property
    def total_iterations(self):
        """
        Gets the total_iterations of this ExperimentStatusExperiments.


        :return: The total_iterations of this ExperimentStatusExperiments.
        :rtype: float
        """
        return self._total_iterations

    @total_iterations.setter
    def total_iterations(self, total_iterations):
        """
        Sets the total_iterations of this ExperimentStatusExperiments.


        :param total_iterations: The total_iterations of this ExperimentStatusExperiments.
        :type: float
        """

        self._total_iterations = total_iterations

    @property
    def state(self):
        """
        Gets the state of this ExperimentStatusExperiments.


        :return: The state of this ExperimentStatusExperiments.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """
        Sets the state of this ExperimentStatusExperiments.


        :param state: The state of this ExperimentStatusExperiments.
        :type: str
        """
        allowed_values = ["running", "completed", "failed", "canceled"]
        if state not in allowed_values:
            raise ValueError(
                "Invalid value for `state` ({0}), must be one of {1}"
                .format(state, allowed_values)
            )

        self._state = state

    @property
    def message(self):
        """
        Gets the message of this ExperimentStatusExperiments.


        :return: The message of this ExperimentStatusExperiments.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this ExperimentStatusExperiments.


        :param message: The message of this ExperimentStatusExperiments.
        :type: str
        """

        self._message = message

    @property
    def failure(self):
        """
        Gets the failure of this ExperimentStatusExperiments.


        :return: The failure of this ExperimentStatusExperiments.
        :rtype: ErrorSchemaExperiments
        """
        return self._failure

    @failure.setter
    def failure(self, failure):
        """
        Sets the failure of this ExperimentStatusExperiments.


        :param failure: The failure of this ExperimentStatusExperiments.
        :type: ErrorSchemaExperiments
        """

        self._failure = failure

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,HW_SPEC_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import HwSpecMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os,json

_DEFAULT_LIST_LENGTH = 50


class HwSpec(WMLResource):
    """
    Store and manage your hardware specs.
    """
    ConfigurationMetaNames = HwSpecMetaNames()
    """MetaNames for Hardware Specification."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, hw_spec_uid):
        """
            Get hardware specification details.

            **Parameters**

            .. important::
                **hw_spec_uid** : Unique id of the hardware spec
                ** type**: str

            **Output**

            .. important::
                **returns**: Metadata of the hardware specifications\n
                **return type**: dict\n

            **Example**

                 >>> hw_spec_details = client.hardware_specifications.get_details(hw_spec_uid)

        """
        HwSpec._validate_type(hw_spec_uid, u'hw_spec_uid', STR_TYPE, True)
        if not self._ICP:
            response = requests.get(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(skip_space_project_chk=True),
                                    headers=self._client._get_headers())
        else:
            if self._client.ICP_PLATFORM_SPACES:
                response = requests.get(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid),
                                        params=self._client._params(skip_space_project_chk=True),
                                        headers=self._client._get_headers(), verify=False)
            else:
                response = requests.get(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(),
                                          headers=self._client._get_headers(), verify=False)

        if response.status_code == 200:
            return self._get_required_element_from_response(self._handle_response(200, u'get hardware spec details', response))
        else:
            return self._handle_response(200, u'get hardware spec details', response)

    # Creation of new hardware specs is not required at the moment because WML does not support it
    # @docstring_parameter({'str_type': STR_TYPE_NAME})
    # def store(self, meta_props):
    #     """
    #             Create a space.
    #
    #             **Parameters**
    #
    #             .. important::
    #                #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
    #                                 >>> client.hardware_specifications.ConfigurationMetaNames.get()
    #
    #                   **type**: dict\n
    #
    #             **Output**
    #
    #             .. important::
    #
    #                 **returns**: metadata of the stored space\n
    #                 **return type**: dict\n
    #
    #             **Example**
    #
    #              >>> meta_props = {
    #              >>>    client.hardware_specifications.ConfigurationMetaNames.NAME: "skl_pipeline_heart_problem_prediction",
    #              >>>    client.hardware_specifications.ConfigurationMetaNames.DESCRIPTION: "description scikit-learn_0.20",
    #              >>>    client.hardware_specifications.ConfigurationMetaNames.HARDWARE_CONFIGURATIONS: {},
    #              >>> }
    #
    #     """
    #
    #     # quick support for COS credentials instead of local path
    #     # TODO add error handling and cleaning (remove the file)
    #     HwSpec._validate_type(meta_props, u'meta_props', dict, True)
    #     hw_spec_meta = self.ConfigurationMetaNames._generate_resource_metadata(
    #         meta_props,
    #         with_validation=True,
    #         client=self._client)
    #
    #     hw_spec_meta_json = json.dumps(hw_spec_meta)
    #     href = self._client.service_instance._href_definitions.get_hw_specs_href()
    #
    #     if not self._ICP:
    #         creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=hw_spec_meta_json)
    #     else:
    #         creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=hw_spec_meta_json, verify=False)
    #
    #     hw_spec_details = self._handle_response(201, u'creating new hw specs', creation_response)
    #
    #     return hw_spec_details


    def list(self, name=None):
        """
           List hardware specifications.

           **Parameters**

           .. important::
                **name** : Unique id of the hardware spec
                ** type**: str

           **Output**

           .. important::
                This method only prints the list of all assets in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.hardware_specifications.list()

        """

        params = {}

        if name is not None:
            params.update({'name': name})

        # Todo provide api to return
        href = self._client.service_instance._href_definitions.get_hw_specs_href()
        # Todo include space_id/project_id once https://github.ibm.com/ax/planning/issues/7493 is fixed
        if not self._ICP:
            response = requests.get(href, params, headers=self._client._get_headers())
        else:
            response = requests.get(href, headers=self._client._get_headers(), verify=False)
        self._handle_response(200, u'list hw_specs', response)
        asset_details = self._handle_response(200, u'list assets', response)["resources"]
        hw_spec_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_id'], m[u'metadata'][u'description']) for
            m in asset_details]
        self._list(hw_spec_values, [u'NAME', u"ID", u'DESCRIPTION'], None, _DEFAULT_LIST_LENGTH)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(hw_spec_details):
        """
                Get ID of hardware specifications asset.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the hardware specifications \n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of hardware specifications\n
                    **return type**: str\n

                **Example**

                 >>> asset_uid = client.hardware_specifications.get_id(hw_spec_details)

        """

        return HwSpec.get_uid(hw_spec_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(hw_spec_details):
        """
                Get UID of hardware specifications asset. Deprecated!! Use get_id(hw_spec_details) instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the hardware specifications \n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of hardware specifications\n
                    **return type**: str\n

                **Example**

                 >>> asset_uid = client.hardware_specifications.get_uid(hw_spec_details)

        """
        HwSpec._validate_type(hw_spec_details, u'hw_spec_details', object, True)
        HwSpec._validate_type_of_details(hw_spec_details, HW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(hw_spec_details, u'hw_spec_details',
                                                           [u'metadata', u'asset_id'])


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(hw_spec_details):
        """
            Get url of hardware specifications.

           **Parameters**

           .. important::
                #. **hw_spec_details**:  hardware specifications details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of hardware specifications\n
                **return type**: str\n

           **Example**

             >>> hw_spec_details = client.hw_spec.get_details(hw_spec_uid)
             >>> hw_spec_href = client.hw_spec.get_href(hw_spec_details)

        """
        HwSpec._validate_type(hw_spec_details, u'hw_spec_details', object, True)
        HwSpec._validate_type_of_details(hw_spec_details, HW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(hw_spec_details, u'hw_spec_details', [u'metadata', u'href'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id_by_name(self, hw_spec_name):
        """
                Get Unique Id of hardware specification for the given name.

                **Parameters**

                .. important::

                   #. **hw_spec_name**:  name of the hardware spec \n
                      **type**: str\n

                **Output**

                .. important::

                    **returns**: Unique Id of hardware specification \n
                    **return type**: str\n

                **Example**

                 >>> asset_uid = client.hardware_specifications.get_id_by_name(hw_spec_name)

        """

        return HwSpec.get_uid_by_name(self, hw_spec_name)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid_by_name(self, hw_spec_name):
        """
                Get Unique Id of hardware specification for the given name. Deprecated!!
                Use get_id_by_name(self, hw_spec_name) instead

                **Parameters**

                .. important::

                   #. **hw_spec_name**:  name of the hardware spec \n
                      **type**: str\n

                **Output**

                .. important::

                    **returns**: Unique Id of hardware specification \n
                    **return type**: str\n

                **Example**

                 >>> asset_uid = client.hardware_specifications.get_uid_by_name(hw_spec_name)

        """
        HwSpec._validate_type(hw_spec_name, u'hw_spec_name', STR_TYPE, True)
        parameters = self._client._params(skip_space_project_chk=True)
        parameters.update(name=hw_spec_name)
        if not self._ICP:
            response = requests.get(self._client.service_instance._href_definitions.get_hw_specs_href(),
                                    params=parameters,
                                    headers=self._client._get_headers())
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_hw_specs_href(),
                                    params=parameters,
                                    headers=self._client._get_headers(), verify=False)
        if response.status_code == 200:
            total_values = self._handle_response(200, u'list assets', response)["total_results"]
            if total_values != 0:
                hw_spec_details = self._handle_response(200, u'list assets', response)["resources"]
                return hw_spec_details[0][u'metadata'][u'asset_id']
            else:
                return "Not Found"

    # @docstring_parameter({'str_type': STR_TYPE_NAME})
    # def delete(self, hw_spec_uid):
    #     """
    #         Delete a hardware specifications.
    #
    #         **Parameters**
    #
    #         .. important::
    #             #. **hw_spec_uid**:  hardware specifications UID\n
    #                **type**: str\n
    #
    #         **Output**
    #
    #         .. important::
    #             **returns**: status ("SUCCESS" or "FAILED")\n
    #             **return type**: str\n
    #
    #         **Example**
    #
    #          >>> client.hw_spec.delete(hw_spec_uid)
    #     """
    #     HwSpec._validate_type(hw_spec_uid, u'hw_spec_uid', STR_TYPE, True)
    #
    #     if not self._ICP:
    #         response = requests.delete(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(),
    #                                 headers=self._client._get_headers())
    #     else:
    #         response = requests.delete(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(),
    #                                   headers=self._client._get_headers(), verify=False)
    #     if response.status_code == 200:
    #         return self._get_required_element_from_response(response.json())
    #     else:
    #         return self._handle_response(204, u'delete hardware specification', response)
    #

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'hw_spec_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:
                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                           },
                              'entity': response_data['entity']

                              }
                else:
                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                       },
                             'entity': response_data['entity']

                            }
            else:
                # For system hardware spec
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})

            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

pkg_name = "ibm-watson-machine-learning"

try:
    from importlib.metadata import version
    version = version(pkg_name)

except (ModuleNotFoundError, AttributeError):
    from importlib_metadata import version as imp_lib_ver
    version = imp_lib_ver(pkg_name)

from ibm_watson_machine_learning.client import APIClient
APIClient.version = version

from .utils import is_python_2
if is_python_2():
    import logging
    logger = logging.getLogger(__name__)
    logger.warning("Python 2 is not officially supported.")


#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from sklearn.pipeline import Pipeline
from pprint import pprint
import pandas as pd
import traceback
from os import environ

from ibm_watson_machine_learning import APIClient

from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.workspace import WorkSpace

from ibm_watson_machine_learning.helpers.connections import S3Connection, S3Location, DataConnection, FSLocation, \
    AssetLocation

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines

from ibm_watson_machine_learning.tests.utils import (get_wml_credentials, get_cos_credentials, bucket_exists,
                                                     create_bucket, get_space_id, is_cp4d)
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment

from ibm_watson_machine_learning.utils.autoai.enums import TShirtSize, PipelineTypes

from lale.operators import TrainablePipeline


@unittest.skip("S3 type no longer supported")
class TestAutoAIRemote(unittest.TestCase):
    """
    The test can be run on CLOUD, WMLS and CPD (not tested)
    The test covers:
    - COS set-up (if run on Cloud): checking if bucket exists for the cos instance, if not new bucket is create
    - Saving data `/bank.cdv` to COS/data assets
    - downloading training data from cos/data assets
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    """
    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    cos_credentials = None
    pipeline_opt: 'RemoteAutoPipelines' = None
    service: 'WebService' = None
    service_batch: 'Batch' = None

    trained_pipeline_details = None
    run_id = None

    data_connection = None
    results_connection = None

    train_data = None

    bucket_name = environ.get('BUCKET_NAME', "wml-autoaitests-qa")
    pod_version = environ.get('KB_VERSION', None)

    cos_endpoint = "https://s3.us-south.cloud-object-storage.appdomain.cloud"
    data_cos_path = 'data/donorchoose_cos.csv'
    cos_resource_instance_id = None

    results_cos_path = 'results_wml_autoai'

    pipeline: 'Pipeline' = None
    lale_pipeline = None
    deployed_pipeline = None
    hyperopt_pipelines = None
    new_pipeline = None
    new_sklearn_pipeline = None
    X_values = None
    y_values = None
    elapsed_time = None
    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')

    data_location = './autoai/data/donorchoose_1MB.csv'

    prediction_column = "project_is_approved"

    with_text_transformer = True

    OPTIMIZER_NAME = f"TEXT TRANSFORMER {with_text_transformer} binary {data_location.split('/')[-1]} "
    DEPLOYMENT_NAME = "TEXT TRANSFORMER AutoAI Deployment tests"

    project_id = None
    space_id = None

    train_sample_columns_index_list = [1, 2, 4, 5, 7,  8, 9]
    text_processing = True
    word2vec_feature_number = 15
    drop_duplicates = True

    def _check_optimizer_parameters(self, parameters):
        # check_pipeline_parameters:
        self.assertIsNotNone(parameters.get('name'))

        self.assertEqual(parameters.get('prediction_column'), self.prediction_column)
        self.assertEqual(parameters.get('text_processing'), self.text_processing)
        self.assertEqual(parameters.get('drop_duplicates'), self.drop_duplicates)
        self.assertEqual(parameters.get('word2vec_feature_number'), self.word2vec_feature_number)

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        if not cls.wml_client.ICP:
            cls.cos_credentials = get_cos_credentials()
            cls.cos_endpoint = cls.cos_credentials.get('endpoint_url')
            cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        cls.project_id = cls.wml_credentials.get('project_id')

    def test_00a_space_cleanup(self):
        space_cleanup(self.wml_client,
                      get_space_id(self.wml_client, self.space_name,
                                   cos_resource_instance_id=self.cos_resource_instance_id),
                      days_old=7)
        TestAutoAIRemote.space_id = get_space_id(self.wml_client, self.space_name,
                                                 cos_resource_instance_id=self.cos_resource_instance_id)
        if self.wml_client.ICP:
            self.wml_client.set.default_project(self.project_id)
        else:
            self.wml_client.set.default_space(self.space_id)

    def test_00b_prepare_COS_instance(self):
        if self.wml_client.ICP:
            self.skipTest("Prepare COS is available only for Cloud")
        import ibm_boto3
        cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(cos_resource, self.bucket_name):
            TestAutoAIRemote.bucket_name = create_bucket(cos_resource, self.bucket_name)

            self.assertIsNotNone(TestAutoAIRemote.bucket_name)
            self.assertTrue(bucket_exists(cos_resource, TestAutoAIRemote.bucket_name))

        print(f"Using COS bucket: {TestAutoAIRemote.bucket_name}")

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        TestAutoAIRemote.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                             project_id=self.project_id)
        # space_id=self.space_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_save_remote_data_and_DataConnection_setup(self):
        if self.wml_client.ICP:
            asset_details = self.wml_client.data_assets.create(
                name=self.data_location.split('/')[-1],
                file_path=self.data_location)
            asset_id = asset_details['metadata']['guid']

            TestAutoAIRemote.data_connection = DataConnection(
                location=AssetLocation(asset_id=asset_id))
            TestAutoAIRemote.results_connection = None

        else:  # for cloud and COS
            TestAutoAIRemote.data_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path)
            )
            TestAutoAIRemote.results_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.results_cos_path)
            )
            TestAutoAIRemote.data_connection.write(data=self.data_location, remote_name=self.data_cos_path)

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)

    def test_03_initialize_optimizer(self):
        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            desc='test description',
            prediction_type=self.experiment.PredictionType.BINARY,
            prediction_column=self.prediction_column,
            scoring=self.experiment.Metrics.LOG_LOSS,
            train_sample_columns_index_list=self.train_sample_columns_index_list,
            text_processing=True,
            word2vec_feature_number=self.word2vec_feature_number,
            autoai_pod_version=self.pod_version
        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_04_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)
        self._check_optimizer_parameters(parameters)
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_05_fit_run_training_of_auto_ai_in_wml(self):
        import time
        start_time = time.time()
        TestAutoAIRemote.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=[self.data_connection],
            training_results_reference=self.results_connection,
            background_mode=False)

        TestAutoAIRemote.run_id = self.trained_pipeline_details['metadata']['id']
        TestAutoAIRemote.elapsed_time = time.time() - start_time
        print(self.elapsed_time)

        self.assertIsNotNone(self.data_connection.auto_pipeline_params,
                             msg='DataConnection auto_pipeline_params was not updated.')
        # TestAutoAIRemote.run_id = '95327361-d68a-4f3b-ba62-1a46d6d34b54'
        # TestAutoAIRemote.remote_auto_pipelines = self.experiment.runs.get_optimizer(self.run_id)

    def test_06_get_train_data(self):
        TestAutoAIRemote.train_data = self.remote_auto_pipelines.get_data_connections()[0].read()

        print("train data sample:")
        print(self.train_data.head())
        self.assertGreater(len(self.train_data), 0)

        TestAutoAIRemote.X_values = self.train_data.drop([self.prediction_column], axis=1)[:10].values
        TestAutoAIRemote.y_values = self.train_data[self.prediction_column][:10]

    def test_07_get_run_status(self):
        status = self.remote_auto_pipelines.get_run_status()
        self.assertEqual(status, "completed", msg="AutoAI run didn't finished successfully. Status: {}".format(status))

    def test_08_get_run_details(self):
        parameters = self.remote_auto_pipelines.get_run_details()
        print(parameters)
        self.assertIsNotNone(parameters)

    def test_08b_get_metrics(self):
        self.wml_client.training.get_metrics(self.run_id)

    def test_09_predict_using_fitted_pipeline(self):
        predictions = self.remote_auto_pipelines.predict(X=self.X_values[:5])
        print(predictions)
        self.assertGreater(len(predictions), 0)

    def test_10_summary_listing_all_pipelines_from_wml(self):
        pipelines_details = self.remote_auto_pipelines.summary()
        print(pipelines_details)

    def test_11__get_data_connections__return_a_list_with_data_connections_with_optimizer_params(self):
        data_connections = self.remote_auto_pipelines.get_data_connections()
        self.assertIsInstance(data_connections, list, msg="There should be a list container returned")
        self.assertIsInstance(data_connections[0], DataConnection,
                              msg="There should be a DataConnection object returned")

    def test_12a_get_pipeline_params_specific_pipeline_parameters(self):
        pipeline_params = self.remote_auto_pipelines.get_pipeline_details(pipeline_name='Pipeline_1')
        print(pipeline_params)

    def test_12a_get_pipeline_params__fetch_best_pipeline_parameters__parameters_fetched_as_dict(self):
        best_pipeline_params = self.remote_auto_pipelines.get_pipeline_details()
        print(best_pipeline_params)

    def test_13_check_params_of_realoaded_pipeline_optimizer(self):
        historical_optimizer = self.experiment.runs.get_optimizer(self.run_id)
        hist_opt_params = historical_optimizer.get_params()
        print(hist_opt_params)
        self._check_optimizer_parameters(hist_opt_params)

    ####
    # LALE #
    ########

    def test_14a_get_pipeline__load_sklearn_pipeline__pipeline_loaded(self):
        TestAutoAIRemote.pipeline = self.remote_auto_pipelines.get_pipeline(pipeline_name='Pipeline_4',
                                                                            astype=PipelineTypes.SKLEARN)
        print(f"Fetched pipeline type: {type(self.pipeline)}")

        self.assertIsInstance(self.pipeline, Pipeline,
                              msg="Fetched pipeline is not of Sklearn Pipeline instance.")

        print(self.pipeline.steps)
        print("\nPredictions:")
        predictions = self.pipeline.predict(
            X=self.X_values[:5])
        print(predictions)

    def test_14b_get_pipeline__load_lale_pipeline__pipeline_loaded(self):
        TestAutoAIRemote.pipeline = self.remote_auto_pipelines.get_pipeline(pipeline_name='Pipeline_4')
        print(f"Fetched pipeline type: {type(self.pipeline)}")

        self.assertIsInstance(self.pipeline, TrainablePipeline,
                              msg="Fetched pipeline is not of TrainablePipeline instance.")
        predictions = self.pipeline.predict(
            X=self.X_values[:5])
        print(predictions)

    def test_15_get_all_pipelines_as_lale(self):
        summary = self.remote_auto_pipelines.summary()
        print(summary)
        failed_pipelines = []
        for pipeline_name in summary.reset_index()['Pipeline Name']:
            print(f"Getting pipeline: {pipeline_name}")
            lale_pipeline = None
            try:
                lale_pipeline = self.remote_auto_pipelines.get_pipeline(pipeline_name=pipeline_name)
                self.assertIsInstance(lale_pipeline, TrainablePipeline)
                print(lale_pipeline)
                print(self.X_values[:5])
                predictions = lale_pipeline.predict(
                    X=self.X_values[:2])
                print(predictions)
                self.assertGreater(len(predictions), 0, msg=f"Returned prediction for {pipeline_name} are empty")
            except:
                print(f"Failure: {pipeline_name}")
                failed_pipelines.append(pipeline_name)
                traceback.print_exc()

            if not TestAutoAIRemote.lale_pipeline:
                TestAutoAIRemote.lale_pipeline = lale_pipeline
                print(f"{pipeline_name} loaded for next test cases")

        self.assertEqual(len(failed_pipelines), 0, msg=f"Some pipelines failed. Full list: {failed_pipelines}")

    # @unittest.skip("Skipped lale pretty print")
    def test_16__pretty_print_lale__checks_if_generated_python_pipeline_code_is_correct(self):
        pipeline_code = self.lale_pipeline.pretty_print()
        try:
            exec(pipeline_code)

        except Exception as exception:
            self.assertIsNone(exception,
                              msg=f"Pretty print from lale pipeline was not successful \n\n Full pipeline code:\n {pipeline_code}")

    #################################
    #      DEPLOYMENT SECTION       #
    #################################

    def test_21_deployment_setup_and_preparation(self):
        TestAutoAIRemote.service = WebService(source_wml_credentials=self.wml_credentials.copy(),
                                              source_project_id=self.project_id,
                                              target_wml_credentials=self.wml_credentials,
                                              target_space_id=self.space_id)
        self.wml_client.set.default_space(self.space_id)
        delete_model_deployment(self.wml_client, deployment_name=self.DEPLOYMENT_NAME)
        self.wml_client.set.default_project(self.project_id) if self.project_id else None

        self.assertIsInstance(self.service, WebService, msg="Deployment is not of WebService type.")
        self.assertIsInstance(self.service._source_workspace, WorkSpace, msg="Workspace set incorrectly.")
        self.assertEqual(self.service.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(self.service.name, None, msg="Deployment name initialized incorrectly")

    def test_22__deploy__deploy_best_computed_pipeline_from_autoai_on_wml(self):
        self.service.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model="Pipeline_1",
            deployment_name=self.DEPLOYMENT_NAME)

        self.assertIsNotNone(self.service.id, msg="Online Deployment creation - missing id")
        self.assertIsNotNone(self.service.name, msg="Online Deployment creation - name not set")
        self.assertIsNotNone(self.service.scoring_url,
                             msg="Online Deployment creation - mscoring url  missing")

    def test_23_score_deployed_model(self):
        nb_records = 5
        predictions = self.service.score(payload=self.train_data.drop([self.prediction_column], axis=1)[:nb_records])
        print(predictions)
        self.assertIsNotNone(predictions)
        self.assertEqual(len(predictions['predictions'][0]['values']), nb_records)

    def test_24_list_deployments(self):
        self.service.list()
        params = self.service.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_25_delete_deployment(self):
        print("Delete current deployment: {}".format(self.service.deployment_id))
        self.service.delete()
        self.wml_client.set.default_space(self.space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(self.service.asset_id)
        self.wml_client.set.default_project(self.project_id) if is_cp4d() else None
        self.assertEqual(self.service.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(self.service.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(self.service.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")

    #########################
    #  Batch deployment
    #########################

    def test_30_batch_deployment_setup_and_preparation(self):
        TestAutoAIRemote.service_batch = Batch(source_wml_credentials=self.wml_credentials.copy(),
                                               source_project_id=self.project_id,
                                               target_wml_credentials=self.wml_credentials,
                                               target_space_id=self.space_id)

        self.assertIsInstance(self.service_batch, Batch, msg="Deployment is not of Batch type.")
        self.assertIsInstance(self.service_batch._source_workspace, WorkSpace, msg="Workspace set incorrectly.")
        self.assertEqual(self.service_batch.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(self.service_batch.name, None, msg="Deployment name initialized incorrectly")

    def test_32__deploy__batch_deploy_best_computed_pipeline_from_autoai_on_wml(self):
        self.service_batch.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model="Pipeline_2",
            deployment_name=self.DEPLOYMENT_NAME + ' BATCH')

        self.assertIsNotNone(self.service_batch.id, msg="Batch Deployment creation - missing id")
        self.assertIsNotNone(self.service_batch.id, msg="Batch Deployment creation - name not set")
        self.assertIsNotNone(self.service_batch.asset_id,
                             msg="Batch Deployment creation - model (asset) id missing, incorrect model storing")

    def test_33_score_batch_deployed_model(self):
        scoring_params = self.service_batch.run_job(
            payload=self.train_data.drop(columns=[self.prediction_column]),
            background_mode=False)
        print(scoring_params)
        print(self.service_batch.get_job_result(scoring_params['metadata']['id']))
        self.assertIsNotNone(scoring_params)
        self.assertIn('predictions', str(scoring_params).lower())

    def test_34_list_batch_deployments(self):
        deployments = self.service_batch.list()
        print(deployments)
        self.assertIn('created_at', str(deployments).lower())
        self.assertIn('status', str(deployments).lower())

        params = self.service_batch.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_35_delete_deployment_batch(self):
        print("Delete current deployment: {}".format(self.service_batch.deployment_id))
        self.service_batch.delete()
        self.wml_client.set.default_space(self.space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(self.service_batch.asset_id)
        self.wml_client.set.default_project(self.project_id) if is_cp4d() else None
        self.assertEqual(self.service_batch.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(self.service_batch.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(self.service_batch.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import os
import uuid

from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d, create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestAutoAIAsync, \
    AbstractTestWebservice, AbstractTestBatch

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, RegressionAlgorithms


class TestAutoAIRemote(AbstractTestAutoAIAsync, AbstractTestWebservice, AbstractTestBatch, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/Hotel_Reviews_1MB.csv'
    data_encoding = "utf_8"

    data_cos_path = 'data/Hotel_Reviews.csv'

    batch_payload_location = './autoai/data/Hotel_Reviews_1MB_scoring_payload.csv'
    batch_payload_cos_location = 'scoring_payload/Hotel_Reviews_1MB_scoring_payload.csv'

    SPACE_ONLY = True

    OPTIMIZER_NAME = "Hotel Reviews text data test sdk"

    BATCH_DEPLOYMENT_WITH_DF = True
    BATCH_DEPLOYMENT_WITH_DA = False
    HISTORICAL_RUNS_CHECK = False

    target_space_id = None


    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='Text Transformer experiment',
        prediction_type=PredictionType.REGRESSION,
        prediction_column='Reviewer_Score',
        scoring=Metrics.ROOT_MEAN_SQUARED_ERROR,
        include_only_estimators=[RegressionAlgorithms.SnapBM, RegressionAlgorithms.XGB],
        max_number_of_estimators=2,
        text_processing=True,
        text_columns_names=['Negative_Review', 'Positive_Review'],
        word2vec_feature_number=5,
        # encoding=data_encoding
    )

    def test_00b_prepare_COS_instance_and_connection(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.data_location,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_00d_prepare_connected_data_asset(self):
        asset_details = self.wml_client.data_assets.store({
            self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.wml_client.data_assets.ConfigurationMetaNames.NAME: "Hotel Reviewers utf16 - training asset",
            self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME:self.data_cos_path
        })
        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.results_connection = None

        # TestAutoAIRemote.results_connection = DataConnection(
        #     connection_asset_id=self.connection_id,
        #     location=S3Location(
        #         bucket=self.bucket_name,
        #         path=self.results_cos_path
        #     )
        # )

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        # self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_08a_get_feature_importance(self):
        feature_importance_df = self.remote_auto_pipelines.get_pipeline_details(pipeline_name='Pipeline_4').get(
            'features_importance')

        self.assertIsNotNone(feature_importance_df)

        str_feature_importance_index_list = str(list(feature_importance_df.sort_index().index))

        text_columns = self.experiment_info['text_columns_names']

        for column in text_columns:
            self.assertIn(f'word2vec({column})', str_feature_importance_index_list,
                          msg=f"word2vec({column}) is not in features importance table. Full table: {feature_importance_df}")

        self.assertIn('NewTextFeature_0', str_feature_importance_index_list,
                      msg="Text features were numerated incorrectly.")
        self.assertIn(f"NewTextFeature_{self.experiment_info['word2vec_feature_number'] * len(text_columns) - 1}",
                      str_feature_importance_index_list, msg="Text features were numerated incorrectly.")


    def test_99_delete_connection_and_connected_data_asset(self):
        if not self.SPACE_ONLY:
            self.wml_client.set.default_project(self.project_id)
        self.wml_client.connections.delete(self.connection_id)
        self.wml_client.data_assets.delete(self.asset_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()
